import enum
import re
from typing import TYPE_CHECKING, Any, Callable, Optional, cast

import fmf
import fmf.utils

import tmt.container
import tmt.identifier
import tmt.log
import tmt.utils
from tmt.checks import CheckEvent, CheckResultInterpret
from tmt.container import SerializableContainer, container, field
from tmt.utils import GeneralError, Path
from tmt.utils.themes import style

if TYPE_CHECKING:
    import tmt.base
    import tmt.steps.execute

# Extra keys used for identification in Result class
EXTRA_RESULT_IDENTIFICATION_KEYS = ['extra-nitrate', 'extra-task']


class ResultOutcome(enum.Enum):
    PASS = 'pass'
    FAIL = 'fail'
    INFO = 'info'
    WARN = 'warn'
    ERROR = 'error'
    SKIP = 'skip'
    PENDING = 'pending'

    @classmethod
    def from_spec(cls, spec: str) -> 'ResultOutcome':
        try:
            return ResultOutcome(spec)
        except ValueError:
            raise tmt.utils.SpecificationError(f"Invalid partial custom result '{spec}'.")

    @staticmethod
    def reduce(outcomes: list['ResultOutcome']) -> 'ResultOutcome':
        """
        Reduce several result outcomes into a single outcome

        Convert multiple outcomes into a single one by picking the
        worst. This is used when aggregating several test or check
        results to present a single value to the user.
        """

        outcomes_by_severity = (
            ResultOutcome.ERROR,
            ResultOutcome.FAIL,
            ResultOutcome.WARN,
            ResultOutcome.PASS,
            ResultOutcome.INFO,
            ResultOutcome.SKIP,
            ResultOutcome.PENDING,
        )

        for outcome in outcomes_by_severity:
            if outcome in outcomes:
                return outcome

        raise GeneralError("No result outcome found to reduce.")


# Cannot subclass enums :/
# https://docs.python.org/3/library/enum.html#restricted-enum-subclassing
class ResultInterpret(enum.Enum):
    # These are "inherited" from ResultOutcome
    PASS = 'pass'
    FAIL = 'fail'
    INFO = 'info'
    WARN = 'warn'
    ERROR = 'error'

    # Special interpret values
    RESPECT = 'respect'
    XFAIL = 'xfail'
    CUSTOM = 'custom'
    RESTRAINT = 'restraint'

    @classmethod
    def is_result_outcome(cls, value: 'ResultInterpret') -> bool:
        return value.name in list(ResultOutcome.__members__.keys())

    @classmethod
    def from_spec(cls, spec: str) -> 'ResultInterpret':
        try:
            return ResultInterpret(spec)
        except ValueError:
            raise tmt.utils.SpecificationError(f"Invalid result interpretation '{spec}'.")

    @classmethod
    def normalize(
        cls,
        key_address: str,
        value: Any,
        logger: tmt.log.Logger,
    ) -> 'ResultInterpret':
        if isinstance(value, ResultInterpret):
            return value

        if isinstance(value, str):
            return cls.from_spec(value)

        raise tmt.utils.SpecificationError(
            f"Invalid result interpretation '{value}' at {key_address}."
        )


RESULT_OUTCOME_COLORS: dict[ResultOutcome, str] = {
    ResultOutcome.PASS: 'green',
    ResultOutcome.FAIL: 'red',
    ResultOutcome.INFO: 'blue',
    ResultOutcome.WARN: 'yellow',
    ResultOutcome.ERROR: 'magenta',
    # TODO (happz) make sure the color is visible for all terminals
    ResultOutcome.SKIP: 'bright_black',
    ResultOutcome.PENDING: 'cyan',
}


#: A type of collection IDs tracked for a single result.
ResultIds = dict[str, Optional[str]]

#: Raw result as written in a YAML file. A dictionary, but for now
#: the actual keys are not important.
RawResult = Any


@container
class ResultGuestData(SerializableContainer):
    """
    Describes what tmt knows about a guest the result was produced on
    """

    name: str = f'{tmt.utils.DEFAULT_NAME}-0'
    role: Optional[str] = None
    primary_address: Optional[str] = None

    @classmethod
    def from_test_invocation(
        cls, *, invocation: 'tmt.steps.execute.TestInvocation'
    ) -> 'ResultGuestData':
        """
        Create a guest data for a result from a test invocation.

        A helper for extracting interesting guest data from a given test
        invocation.

        :param invocation: a test invocation capturing the test run and results.
        """

        return ResultGuestData(
            name=invocation.guest.name,
            role=invocation.guest.role,
            primary_address=invocation.guest.primary_address,
        )


# This needs to be a stand-alone function because of the import of `tmt.base`.
# It cannot be imported on module level because of circular dependency.
def _unserialize_fmf_id(serialized: 'tmt.base._RawFmfId') -> 'tmt.base.FmfId':
    from tmt.base import FmfId

    return FmfId.from_spec(serialized)


@container
class BaseResult(SerializableContainer):
    """
    Describes what tmt knows about a result
    """

    name: str
    result: ResultOutcome = field(
        default=ResultOutcome.PASS,
        serialize=lambda result: result.value,
        unserialize=ResultOutcome.from_spec,
    )
    original_result: ResultOutcome = field(
        default=ResultOutcome.PASS,
        serialize=lambda result: result.value,
        unserialize=ResultOutcome.from_spec,
    )
    note: list[str] = field(
        default_factory=cast(Callable[[], list[str]], list),
        unserialize=lambda value: [] if value is None else value,
    )
    log: list[Path] = field(
        default_factory=cast(Callable[[], list[Path]], list),
        serialize=lambda logs: [str(log) for log in logs],
        unserialize=lambda value: [Path(log) for log in value],
    )

    start_time: Optional[str] = None
    end_time: Optional[str] = None
    duration: Optional[str] = None

    def __post_init__(self) -> None:
        self.original_result = self.result

    def show(self) -> str:
        """
        Return a nicely colored result with test name (and note)
        """

        result = 'errr' if self.result == ResultOutcome.ERROR else self.result.value

        components: list[str] = [
            style(result, fg=RESULT_OUTCOME_COLORS[self.result]),
            self.name,
        ]

        if self.note:
            components.append(f'({self.printable_note})')

        return ' '.join(components)

    @property
    def printable_note(self) -> str:
        return ', '.join(self.note)


@container
class CheckResult(BaseResult):
    """
    Describes what tmt knows about a single test check result
    """

    event: CheckEvent = field(
        default=CheckEvent.BEFORE_TEST,
        serialize=lambda event: event.value,
        unserialize=CheckEvent.from_spec,
    )

    def to_subcheck(self) -> 'SubCheckResult':
        """
        Convert check to a tmt SubCheckResult
        """

        return SubCheckResult.from_serialized(self.to_serialized())


@container
class SubCheckResult(CheckResult):
    """
    Describes what tmt knows about a single subtest check result.

    It does not contain any additional fields; it simply defines a type to
    easily differentiate between a :py:class:`tmt.result.CheckResult` and a
    ``CheckResult`` located within a result phase.
    """


@container
class SubResult(BaseResult):
    """
    Describes what tmt knows about a single test subresult
    """

    check: list[SubCheckResult] = field(
        default_factory=cast(Callable[[], list[SubCheckResult]], list),
        serialize=lambda results: [result.to_serialized() for result in results],
        unserialize=lambda serialized: [
            SubCheckResult.from_serialized(check) for check in serialized
        ],
    )

    @staticmethod
    def failures(log: Optional[str], msg_type: str = 'FAIL') -> str:
        """
        Filter stdout and get only messages with certain type
        """
        return Result.failures(log, msg_type)


@container
class PhaseResult(BaseResult):
    """
    Describes what tmt knows about result of individual phases, e.g. prepare ansible
    """


@container
class Result(BaseResult):
    """
    Describes what tmt knows about a single test result
    """

    serial_number: int = 0
    fmf_id: Optional['tmt.base.FmfId'] = field(
        default=cast(Optional['tmt.base.FmfId'], None),
        serialize=lambda fmf_id: fmf_id.to_minimal_spec() if fmf_id is not None else {},
        unserialize=_unserialize_fmf_id,
    )
    context: tmt.utils.FmfContext = field(
        default_factory=tmt.utils.FmfContext,
        serialize=lambda context: context.to_spec(),
        unserialize=lambda serialized: tmt.utils.FmfContext(serialized),
    )
    ids: ResultIds = field(default_factory=cast(Callable[[], ResultIds], dict))
    guest: ResultGuestData = field(
        default_factory=ResultGuestData,
        serialize=lambda value: value.to_serialized(),
        unserialize=lambda serialized: ResultGuestData.from_serialized(serialized),
    )

    subresult: list[SubResult] = field(
        default_factory=cast(Callable[[], list[SubResult]], list),
        serialize=lambda results: [result.to_serialized() for result in results],
        unserialize=lambda serialized: [
            SubResult.from_serialized(subresult) for subresult in serialized
        ],
    )

    check: list[CheckResult] = field(
        default_factory=cast(Callable[[], list[CheckResult]], list),
        serialize=lambda results: [result.to_serialized() for result in results],
        unserialize=lambda serialized: [
            CheckResult.from_serialized(check) for check in serialized
        ],
    )
    data_path: Optional[Path] = field(
        default=cast(Optional[Path], None),
        serialize=lambda path: None if path is None else str(path),
        unserialize=lambda value: None if value is None else Path(value),
    )

    @classmethod
    def from_test_invocation(
        cls,
        *,
        invocation: 'tmt.steps.execute.TestInvocation',
        result: ResultOutcome,
        note: Optional[list[str]] = None,
        ids: Optional[ResultIds] = None,
        log: Optional[list[Path]] = None,
        subresult: Optional[list[SubResult]] = None,
    ) -> 'Result':
        """
        Create a result from a test invocation.

        A helper for extracting interesting data from a given test invocation.
        While it's perfectly possible to go directly through ``Result(...)``,
        most of the time a result stems from a particular test invocation
        captured by a :py:class:`TestInvocation` instance.

        :param invocation: a test invocation capturing the test run and results.
        :param result: actual test outcome. It will be interpreted according to
            :py:attr:`Test.result` key (see
            https://tmt.readthedocs.io/en/stable/spec/tests.html#result).
        :param note: optional result notes.
        :param ids: additional test IDs. They will be added to IDs extracted
            from the test.
        :param log: optional list of test logs.
        """

        # Saving identifiable information for each test case so we can match them
        # to Polarion/Nitrate/other cases and report run results there
        # TODO: would an exception be better? Can test.id be None?
        ids = ids or {}
        default_ids: ResultIds = {tmt.identifier.ID_KEY: invocation.test.id}

        for key in EXTRA_RESULT_IDENTIFICATION_KEYS:
            value: Any = cast(Any, invocation.test.node.get(key))

            default_ids[key] = None if value is None else str(value)

        default_ids.update(ids)
        ids = default_ids

        _result = Result(
            name=invocation.test.name,
            serial_number=invocation.test.serial_number,
            fmf_id=invocation.test.fmf_id,
            context=invocation.phase.step.plan._fmf_context,
            result=result,
            note=note or [],
            start_time=invocation.start_time,
            end_time=invocation.end_time,
            duration=invocation.real_duration,
            ids=ids,
            log=log or [],
            guest=ResultGuestData.from_test_invocation(invocation=invocation),
            data_path=invocation.relative_test_data_path,
            subresult=subresult or [],
            check=invocation.check_results or [],
        )

        interpret_checks = {check.how: check.result for check in invocation.test.check}

        return _result.interpret_result(invocation.test.result, interpret_checks)

    def interpret_check_result(
        self,
        check_name: str,
        interpret_checks: dict[str, CheckResultInterpret],
    ) -> ResultOutcome:
        """
        Aggregate all checks of given name and interpret the outcome

        :param check_name: name of the check to be aggregated
        :param interpret_checks: mapping of check:how and its result interpret
        :returns: :py:class:`ResultOutcome` instance with the interpreted result
        """

        # Reduce all check outcomes into a single worst outcome
        reduced_outcome = ResultOutcome.reduce(
            [check.result for check in self.check if check.name == check_name]
        )

        # Now let's handle the interpretation
        interpret = interpret_checks[check_name]
        interpreted_outcome = reduced_outcome

        if interpret == CheckResultInterpret.RESPECT:
            if interpreted_outcome == ResultOutcome.FAIL:
                self.note.append(f"check '{check_name}' failed")

        elif interpret == CheckResultInterpret.INFO:
            interpreted_outcome = ResultOutcome.INFO
            self.note.append(f"check '{check_name}' is informational")

        elif interpret == CheckResultInterpret.XFAIL:
            if reduced_outcome == ResultOutcome.PASS:
                interpreted_outcome = ResultOutcome.FAIL
                self.note.append(f"check '{check_name}' did not fail as expected")

            if reduced_outcome == ResultOutcome.FAIL:
                interpreted_outcome = ResultOutcome.PASS
                self.note.append(f"check '{check_name}' failed as expected")

        return interpreted_outcome

    def interpret_result(
        self,
        interpret: ResultInterpret,
        interpret_checks: dict[str, CheckResultInterpret],
    ) -> 'Result':
        """
        Interpret result according to a given interpretation instruction.

        Inspect and possibly modify :py:attr:`result` and :py:attr:`note`
        attributes, following the ``interpret`` value.

        :param interpret: how to interpret current result.
        :param interpret_checks: mapping of check:how and its result interpret
        :returns: :py:class:`Result` instance containing the updated result.
        """

        if interpret not in ResultInterpret:
            raise tmt.utils.SpecificationError(
                f"Invalid result '{interpret.value}' in test '{self.name}'."
            )

        if interpret == ResultInterpret.CUSTOM:
            return self

        # Interpret check results (aggregated by the check name)
        check_outcomes: list[ResultOutcome] = []
        for check_name in tmt.utils.uniq([check.name for check in self.check]):
            check_outcomes.append(self.interpret_check_result(check_name, interpret_checks))

        # Aggregate check results with the main test result
        self.result = ResultOutcome.reduce([self.result, *check_outcomes])

        # Override result with result outcome provided by user
        if interpret not in (ResultInterpret.RESPECT, ResultInterpret.XFAIL):
            self.result = ResultOutcome(interpret.value)
            self.note.append(f"test result overridden: {self.result.value}")

            # Add original result to note if the result has changed
            if self.result != self.original_result:
                self.note.append(f"original test result: {self.original_result.value}")

            return self

        # Handle the expected fail
        if interpret == ResultInterpret.XFAIL:
            if self.result == ResultOutcome.PASS:
                self.result = ResultOutcome.FAIL
                self.note.append("test was expected to fail")

            elif self.result == ResultOutcome.FAIL:
                self.result = ResultOutcome.PASS
                self.note.append("test failed as expected")

        # Add original result to note if the result has changed
        if self.result != self.original_result:
            self.note.append(f"original test result: {self.original_result.value}")

        return self

    def to_subresult(self) -> 'SubResult':
        """
        Convert result to tmt subresult
        """
        options = [
            tmt.container.key_to_option(key) for key in tmt.container.container_keys(SubResult)
        ]

        return SubResult.from_serialized(
            {option: value for option, value in self.to_serialized().items() if option in options}
        )

    @staticmethod
    def total(results: list['Result']) -> dict[ResultOutcome, int]:
        """
        Return dictionary with total stats for given results
        """

        stats = dict.fromkeys(RESULT_OUTCOME_COLORS, 0)

        for result in results:
            stats[result.result] += 1
        return stats

    @staticmethod
    def summary(results: list['Result']) -> str:
        """
        Prepare a nice human summary of provided results
        """

        stats = Result.total(results)
        comments = []
        if stats.get(ResultOutcome.PASS):
            passed = ' ' + style('passed', fg='green')
            comments.append(fmf.utils.listed(stats[ResultOutcome.PASS], 'test') + passed)
        if stats.get(ResultOutcome.FAIL):
            failed = ' ' + style('failed', fg='red')
            comments.append(fmf.utils.listed(stats[ResultOutcome.FAIL], 'test') + failed)
        if stats.get(ResultOutcome.SKIP):
            skipped = ' ' + style('skipped', fg='bright_black')
            comments.append(fmf.utils.listed(stats[ResultOutcome.SKIP], 'test') + skipped)
        if stats.get(ResultOutcome.INFO):
            count, comment = fmf.utils.listed(stats[ResultOutcome.INFO], 'info').split()
            comments.append(count + ' ' + style(comment, fg='blue'))
        if stats.get(ResultOutcome.WARN):
            count, comment = fmf.utils.listed(stats[ResultOutcome.WARN], 'warn').split()
            comments.append(count + ' ' + style(comment, fg='yellow'))
        if stats.get(ResultOutcome.ERROR):
            count, comment = fmf.utils.listed(stats[ResultOutcome.ERROR], 'error').split()
            comments.append(count + ' ' + style(comment, fg='magenta'))
        if stats.get(ResultOutcome.PENDING):
            count, comment = str(stats[ResultOutcome.PENDING]), 'pending'
            comments.append(count + ' ' + style(comment, fg='cyan'))
        # FIXME: cast() - https://github.com/teemtee/fmf/issues/185
        return cast(str, fmf.utils.listed(comments or ['no results found']))

    def show(self, display_guest: bool = True) -> str:
        """
        Return a nicely colored result with test name (and note)
        """

        from tmt.steps.provision import format_guest_full_name

        result = 'errr' if self.result == ResultOutcome.ERROR else self.result.value

        components: list[str] = [
            style(result, fg=RESULT_OUTCOME_COLORS[self.result]),
            self.name,
        ]

        if display_guest and self.guest:
            assert self.guest.name  # narrow type

            components.append(f'(on {format_guest_full_name(self.guest.name, self.guest.role)})')

        if self.note:
            components.append(f'({self.printable_note})')

        return ' '.join(components)

    @staticmethod
    def failures(log: Optional[str], msg_type: str = 'FAIL') -> str:
        """
        Filter stdout and get only messages with certain type
        """

        if not log:
            return ''
        filtered = ''

        # Filter beakerlib style logs in the following way:
        # 1. Reverse the log string by lines
        # 2. Search for each FAIL and extract every associated line.
        # 3. For failed phases also extract phase name so the log is easier to understand
        # 4. Reverse extracted lines back into correct order.
        if re.search(':: \\[   FAIL   \\] ::', log):  # dumb check for a beakerlib log
            copy_line = False
            copy_phase_name = False
            failure_log: list[str] = []
            # we will be processing log lines in a reversed order
            iterator = iter(reversed(log.split("\n")))
            for line in iterator:
                # found FAIL enables log extraction
                if re.search(':: \\[   FAIL   \\] ::', line):
                    copy_line = True
                    copy_phase_name = True
                # BEGIN of rlRun block or previous command or beginning of a test section
                # disables extraction
                elif re.search('(:: \\[.{10}\\] ::|[:]{80})', line):
                    copy_line = False
                # extract line from the log
                if copy_line:
                    failure_log.append(line)
                # Add beakerlib phase name to a failure log, in order to properly match the phase
                # name we need to do this in two steps.
                if copy_phase_name and re.search('[:]{80}', line):
                    # read the next line containing phase name
                    line = next(iterator)
                    failure_log.append(f'\n{line}')
                    copy_phase_name = False
            # reverse extracted lines to restore previous order
            failure_log.reverse()
            return '\n'.join(failure_log).strip()

        # Check for other failures and errors when not using beakerlib
        for m in re.findall(
            rf'.*\b(?=error|fail|{msg_type})\b.*', log, re.IGNORECASE | re.MULTILINE
        ):
            filtered += m + '\n'

        return filtered or log


def results_to_exit_code(results: list[Result], execute_enabled: bool = True) -> int:
    """
    Map results to a tmt exit code
    """

    from tmt.cli import TmtExitCode

    stats = Result.total(results)

    # Quoting the specification:

    # "No test results found."
    if sum(stats.values()) == 0:
        return TmtExitCode.NO_RESULTS_FOUND

    # "Errors occurred during test execution."
    if stats[ResultOutcome.ERROR]:
        return TmtExitCode.ERROR

    # "There was a fail or warn identified, but no error."
    if stats[ResultOutcome.FAIL] + stats[ResultOutcome.WARN]:
        return TmtExitCode.FAIL

    # "Tests were executed, and all reported the ``skip`` result."
    if sum(stats.values()) == stats[ResultOutcome.SKIP]:
        return TmtExitCode.ALL_TESTS_SKIPPED

    # "No errors or fails, but there are pending tests."
    if execute_enabled and stats[ResultOutcome.PENDING]:
        return TmtExitCode.ERROR

    # "At least one test passed, there was no fail, warn or error."
    if (
        sum(stats.values())
        == stats[ResultOutcome.PASS]
        + stats[ResultOutcome.INFO]
        + stats[ResultOutcome.SKIP]
        + stats[ResultOutcome.PENDING]
    ):
        return TmtExitCode.SUCCESS

    raise GeneralError("Unhandled combination of test result.")



import re
from typing import Optional, Union

import tmt.utils
from tmt.package_managers import (
    FileSystemPath,
    Installable,
    Options,
    Package,
    PackageManager,
    PackageManagerEngine,
    PackagePath,
    escape_installables,
    provides_package_manager,
)
from tmt.utils import (
    Command,
    CommandOutput,
    GeneralError,
    RunError,
    ShellScript,
)
from tmt.utils.templates import render_template

ReducedPackages = list[Union[Package, PackagePath]]


PRESENCE_TEMPLATE = """
set -x

export DEBIAN_FRONTEND=noninteractive

{% for installable in REAL_PACKAGES %}
echo "PRESENCE-TEST:{{ installable }}:{{ installable }}:$(dpkg-query --show {{ installable }})"
{% endfor %}

{% if FILESYSTEM_PATHS %}
  {% for installable in FILESYSTEM_PATHS %}
fs_path_package="$(apt-file search --package-only {{ installable }})"
if [ -z "$fs_path_package" ]; then
    echo "PRESENCE-TEST:{{ installable }}::"
else
    echo "PRESENCE-TEST:{{ installable }}:${fs_path_package}:$(dpkg-query --show $fs_path_package)"
fi
  {% endfor %}
{% endif %}
"""


INSTALL_TEMPLATE = """
set -x

export DEBIAN_FRONTEND=noninteractive

installable_packages="{{ REAL_PACKAGES | join(' ') }}"

{% if FILESYSTEM_PATHS %}
  {% for installable in FILESYSTEM_PATHS %}
fs_path_package="$(apt-file search --package-only {{ installable }})"
    {% if not OPTIONS.skip_missing %}
[ -z "$fs_path_package" ] && echo "No package found for path {{ installable }}" && exit 1
    {% endif %}
installable_packages="$installable_packages $fs_path_package"
  {% endfor %}
{% endif %}

{% if OPTIONS.check_first %}
dpkg-query --show $installable_packages \\
{% else -%}
/bin/false \\
{% endif -%}
{{ '||' if COMMAND == 'install' else '&&' }} {{ ENGINE.command.to_script() }} {{ COMMAND }} {{ ENGINE.options.to_script() }} {{ EXTRA_OPTIONS }} $installable_packages

{% if OPTIONS.skip_missing %}
exit 0
{% else %}
exit $?
{% endif %}
"""  # noqa: E501


class AptEngine(PackageManagerEngine):
    install_command = Command('install')

    _sudo_prefix: Command

    def prepare_command(self) -> tuple[Command, Command]:
        """
        Prepare installation command for apt
        """

        if self.guest.facts.is_superuser is False:
            self._sudo_prefix = Command('sudo')

        else:
            self._sudo_prefix = Command()

        command = Command()
        options = Command('-y')

        command += self._sudo_prefix
        command += Command('apt')

        return (command, options)

    def _enable_apt_file(self) -> ShellScript:
        return ShellScript(
            f'( {self.install(Package("apt-file"))} ) && {self._sudo_prefix} apt-file update'
        )

    def _reduce_to_packages(
        self, *installables: Installable
    ) -> tuple[list[Union[Package, PackagePath]], list[FileSystemPath]]:
        real_packages: list[Union[Package, PackagePath]] = []
        filesystem_paths: list[FileSystemPath] = []

        for installable in installables:
            if isinstance(installable, (Package, PackagePath)):
                real_packages.append(installable)

            elif isinstance(installable, FileSystemPath):
                filesystem_paths.append(installable)

            else:
                raise GeneralError(f"Package specification '{installable}' is not supported.")

        return (real_packages, filesystem_paths)

    def check_presence(self, *installables: Installable) -> ShellScript:
        real_packages, filesystem_paths = self._reduce_to_packages(*installables)

        if filesystem_paths:
            return self._enable_apt_file() + ShellScript(
                render_template(
                    PRESENCE_TEMPLATE,
                    ENGINE=self,
                    REAL_PACKAGES=escape_installables(*real_packages),
                    FILESYSTEM_PATHS=escape_installables(*filesystem_paths),
                )
            )

        return ShellScript(
            render_template(
                PRESENCE_TEMPLATE,
                ENGINE=self,
                REAL_PACKAGES=escape_installables(*real_packages),
                FILESYSTEM_PATHS=escape_installables(*filesystem_paths),
            )
        )

    def _extra_options(self, options: Options) -> Command:
        extra_options = Command()

        if options.skip_missing:
            extra_options += Command('--ignore-missing')

        return extra_options

    def refresh_metadata(self) -> ShellScript:
        return ShellScript(
            f'export DEBIAN_FRONTEND=noninteractive; {self.command.to_script()} update'
        )

    def install(
        self,
        *installables: Installable,
        options: Optional[Options] = None,
    ) -> ShellScript:
        options = options or Options()
        extra_options = self._extra_options(options)

        real_packages, filesystem_paths = self._reduce_to_packages(*installables)

        if filesystem_paths:
            return self._enable_apt_file() & ShellScript(
                render_template(
                    INSTALL_TEMPLATE,
                    ENGINE=self,
                    COMMAND='install',
                    OPTIONS=options,
                    EXTRA_OPTIONS=extra_options,
                    REAL_PACKAGES=escape_installables(*real_packages),
                    FILESYSTEM_PATHS=escape_installables(*filesystem_paths),
                )
            )

        return ShellScript(
            render_template(
                INSTALL_TEMPLATE,
                ENGINE=self,
                COMMAND='install',
                OPTIONS=options,
                EXTRA_OPTIONS=extra_options,
                REAL_PACKAGES=escape_installables(*real_packages),
                FILESYSTEM_PATHS=escape_installables(*filesystem_paths),
            )
        )

    def reinstall(
        self,
        *installables: Installable,
        options: Optional[Options] = None,
    ) -> ShellScript:
        options = options or Options()
        extra_options = self._extra_options(options)

        real_packages, filesystem_paths = self._reduce_to_packages(*installables)

        if filesystem_paths:
            return self._enable_apt_file() & ShellScript(
                render_template(
                    INSTALL_TEMPLATE,
                    ENGINE=self,
                    COMMAND='reinstall',
                    OPTIONS=options,
                    EXTRA_OPTIONS=extra_options,
                    REAL_PACKAGES=escape_installables(*real_packages),
                    FILESYSTEM_PATHS=escape_installables(*filesystem_paths),
                )
            )

        return ShellScript(
            render_template(
                INSTALL_TEMPLATE,
                ENGINE=self,
                COMMAND='reinstall',
                OPTIONS=options,
                EXTRA_OPTIONS=extra_options,
                REAL_PACKAGES=real_packages,
                FILESYSTEM_PATHS=filesystem_paths,
            )
        )

    def install_debuginfo(
        self,
        *installables: Installable,
        options: Optional[Options] = None,
    ) -> ShellScript:
        raise tmt.utils.GeneralError("There is no support for debuginfo packages in apt.")


@provides_package_manager('apt')
class Apt(PackageManager[AptEngine]):
    NAME = 'apt'

    _engine_class = AptEngine

    probe_command = Command('apt', '--version')

    def check_presence(self, *installables: Installable) -> dict[Installable, bool]:
        presence_script = self.engine.check_presence(*installables)

        try:
            output = self.guest.execute(presence_script)
            stdout, stderr = output.stdout, output.stderr

        except RunError as exc:
            stdout, stderr = exc.stdout, exc.stderr

        if stdout is None or stderr is None:
            raise GeneralError("apt presence check provided no output")

        results: dict[Installable, bool] = {}

        for installable in installables:
            results[installable] = False

            match = re.search(
                rf'(?m)^PRESENCE-TEST:{"".join(escape_installables(installable))}:.*?:(.*?)$',
                stdout,
            )

            if match is None:
                continue

            if match.group(1):
                results[installable] = True

        return results

    def install_debuginfo(
        self,
        *installables: Installable,
        options: Optional[Options] = None,
    ) -> CommandOutput:
        raise tmt.utils.GeneralError("There is no support for debuginfo packages in apt.")



from typing import Any

import pytest

import tmt.cli.trying


@pytest.mark.parametrize(
    ('params', 'expected'),
    [
        (
            {'image_and_how': ('fedora@virtual',), 'arch': None},
            {'image': 'fedora', 'how': 'virtual'},
        ),
        (
            {'image_and_how': ('fedora@virtual',), 'arch': 'aarch64'},
            {'image': 'fedora', 'how': 'virtual', 'arch': 'aarch64'},
        ),
        (
            {'image_and_how': (), 'arch': 'aarch64'},
            {'arch': 'aarch64'},
        ),
    ],
)
def test_options_arch(params: dict[str, Any], expected: dict[str, Any]):
    assert tmt.cli.trying._construct_trying_provision_options(params) == expected



import re

import tmt
import tmt.utils

logger = tmt.Logger.create()
tree = tmt.Tree(path='data', logger=logger)

prefix = r'https://github.com/.*/tmt/tree/.*/tests/core/web-link/data/'


def test_stories():
    url = tree.stories(names=['/story'])[0].web_link()
    assert re.match(prefix + 'story.fmf', url)


def test_plans():
    url = tree.plans(names=['/plan'])[0].web_link()
    assert re.match(prefix + 'plan.fmf', url)


def test_tests():
    url = tree.tests(names=['/test'])[0].web_link()
    assert re.match(prefix + 'test.fmf', url)



"""
Test Management Tool
"""

import importlib.metadata

__version__ = importlib.metadata.version(__name__)

__all__ = [
    'Clean',
    'Guest',
    'GuestSsh',
    'Logger',
    'Plan',
    'Result',
    'Run',
    'Status',
    'Story',
    'Test',
    'Tree',
]

from tmt.base import Clean, Plan, Run, Status, Story, Test, Tree
from tmt.log import Logger
from tmt.result import Result
from tmt.steps.provision import Guest, GuestSsh



"""
Template rendering.

Package provides primitives for Jinja2 template rendering, plus our
custom filters.
"""

import re
import shlex
import textwrap
from re import Match
from typing import (
    Any,
    Callable,
    Optional,
    cast,
)

import fmf
import fmf.utils
import jinja2
import jinja2.exceptions

from tmt.utils import GeneralError, Path
from tmt.utils.git import web_git_url


def _template_filter_basename(  # type: ignore[reportUnusedFunction,unused-ignore]
    pathlike: str,
) -> str:
    """
    Return the last component of the given path.

    .. code-block:: jinja

        # "/etc/fstab" -> "fstab"
        {{ "/etc/fstab" | basename }}

        # "/var/log/" -> "log"
        {{ "/var/log/" | basename }}
    """

    return Path(pathlike).name


def _template_filter_match(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str, pattern: str
) -> Optional[Match[str]]:
    """
    Return `re.Match`__ if the string matches a given pattern.

    Pattern is tested in "match" mode, i.e. it must match from the
    beginning of the string. See :ref:`regular-expressions` description
    for more details.

    __ https://docs.python.org/3.9/library/re.html#match-objects

    .. code-block:: jinja

         # 'foo/bar' -> 'foo/bar'
        {{ 'foo/bar' | match('foo/.*').group() }}

        # 'foo/bar' -> ''
        {{ 'foo/bar' | match('foo/(.+?)/(.*)') }}

        # 'foo/bar/baz' -> 'bar'
        {{ 'foo/bar' | match('foo/(.+?)/.*').group(1) }}
    """

    return re.match(pattern, s)


def _template_filter_search(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str, pattern: str
) -> Optional[Match[str]]:
    """
    Return `re.Match`__ if the string matches a given pattern.

    Pattern is tested in "search" mode, i.e. it can match anywhere
    in the string. See :ref:`regular-expressions` description for more
    details.

    __ https://docs.python.org/3.9/library/re.html#match-objects

    .. code-block:: jinja

         # 'baz/foo/bar' -> 'foo/bar'
        {{ 'baz/foo/bar' | search('foo/.*').group() }}

        # 'baz/foo/bar' -> ''
        {{ 'baz/foo/bar' | search('foo/(.+?)/(.*)') }}

        # 'baz/foo/bar/baz' -> 'bar'
        {{ 'baz/foo/bar' | search('foo/(.+?)/.*').group(1) }}
    """

    return re.search(pattern, s)


def _template_filter_regex_findall(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str, pattern: str
) -> list[str]:
    """
    Return a list of all non-overlapping matches in the string.

    If one or more capturing groups are present in the pattern, return
    a list of groups; this will be a list of tuples if the pattern
    has more than one group.

    Empty matches are included in the result.

    .. code-block:: jinja

        # '/var/log/mail.log' => ['/', '/', '/']
        {{ '/var/log/mail.log' | regex_findall('/') }}
    """

    return re.findall(pattern, s)


def _template_filter_regex_match(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str, pattern: str
) -> str:
    """
    Return string matching a given pattern.

    Pattern is tested in "match" mode, i.e. it must match from the
    beginning of the string. See :ref:`regular-expressions` description
    for more details.

    If the string matches and pattern contains capture groups, the
    first group is returned. If the string matches, but patterns
    contains no capture group, the whole match is returned.
    Otherwise, an empty string is returned.

    .. code-block:: jinja

        # 'foo/bar' -> 'foo/bar'
        {{ 'foo/bar' | regex_match('foo/.*') }}

        # 'foo/bar' -> ''
        {{ 'foo/bar' | regex_match('foo/(.+?)/(.*)') }}

        # 'foo/bar/baz' -> 'bar'
        {{ 'foo/bar/baz' | regex_match('foo/(.+?)/.*') }}
    """

    match = re.match(pattern, s)

    if match is None:
        return ''

    if not match.groups():
        return match.group()

    return match.group(1)


def _template_filter_regex_search(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str, pattern: str
) -> str:
    """
    Return string matching a given pattern.

    Pattern is tested in "search" mode, i.e. it can match anywhere
    in the string. See :ref:`regular-expressions` description for more
    details.

    If the string matches and pattern contains capture groups, the
    first group is returned. If the string matches, but patterns
    contains no capture group, the whole match is returned.
    Otherwise, an empty string is returned.

    .. code-block:: jinja

        # 'baz/foo/bar' -> 'foo/bar'
        {{ 'baz/foo/bar' | regex_search('foo/.*') }}

        # 'baz/foo/bar' -> ''
        {{ 'baz/foo/bar' | regex_search('foo/(.+?)/(.*)') }}

        # 'baz/foo/bar/baz' -> 'bar'
        {{ 'baz/foo/bar/baz' | regex_search('foo/(.+?)/.*') }}
    """

    match = re.search(pattern, s)

    if match is None:
        return ''

    if not match.groups():
        return match.group()

    return match.group(1)


def _template_filter_regex_replace(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str, pattern: str, repl: str
) -> str:
    """
    Replace a substring defined by a regular expression with another string.

    Return the string obtained by replacing the leftmost
    non-overlapping occurrences of pattern in string by the replacement.
    If the pattern isn't found, string is returned unchanged.

    Backreferences in the replacement string are replaced with the
    substring matched by a group in the pattern.

    .. code-block:: jinja

        # 'foo/bar' -> 'foo/baz'
        {{ 'foo/bar' | regex_replace('(.+)/bar', '\1/baz') }}

        # 'foo/bar' -> 'foo/bar'
        {{ 'foo/bar' | regex_replace('(.+)/baz', '\1/') }}
    """

    return re.sub(pattern, repl, s)


def _template_filter_dedent(s: str) -> str:  # type: ignore[reportUnusedFunction,unused-ignore]
    """
    Remove any common leading whitespace from every line in the string.

    .. code-block:: jinja

        # "foo bar" -> "foo bar"
        {{ "foo bar" | dedent }}

        # '''
        #    foo
        #    bar
        #        baz
        # '''
        #  ->
        # '''
        # foo
        # bar
        #    baz
        # '''
        {{ "\\n    foo\\n    bar\\n        baz\\n" | dedent }}
    """

    return textwrap.dedent(s)


def _template_filter_listed(  # type: ignore[reportUnusedFunction,unused-ignore]
    items: list[Any],
    singular: Optional[str] = None,
    plural: Optional[str] = None,
    max: Optional[int] = None,
    quote: str = "",
    join: str = "and",
) -> str:
    """
    Return a nice, human readable description of an iterable.

    .. code-block:: jinja

        # [0] -> "0"
        {{ [0] | listed }}

        # [0, 1] -> "0 and 1"
        {{ [0, 1] | listed }}

        # [0, 1, 2] -> "0, 1, or 2"
        {{ [0, 1, 2] | listed(join='or') }}

        # [0, 1, 2] -> '"0", "1" and "2"'
        {{ [0, 1, 2] | listed(quote='"') }}

        # [0, 1, 2, 3] -> "0, 1, 2 and 1 more"
        {{ [0, 1, 2, 3] | listed(max=3) }}

        # [0, 1, 2, 3, 4] -> "0, 1, 2 and 2 more numbers"
        {{ [0, 1, 2, 3, 4, 5] | listed('number', max=3) }}

        # [0, 1, 2, 3, 4, 5] -> "6 categories"
        {{ [0, 1, 2, 3, 4, 5] | listed('category') }}

        # [0, 1, 2, 3, 4, 5, 6] -> "7 leaves"
        {{ [0, 1, 2, 3, 4, 5, 6] | listed("leaf", "leaves") }}
    """

    # cast: for some reason, mypy sees `listed` as `Any`
    return cast(
        str,
        fmf.utils.listed(items, singular=singular, plural=plural, max=max, quote=quote, join=join),
    )


def _template_filter_web_git_url(  # type: ignore[reportUnusedFunction,unused-ignore]
    path_str: str,
    url: str,
    ref: str,
) -> str:
    """
    Sanitize git url using :py:meth:`tmt.utils.web_git_url`

    .. code-block:: jinja

        {{ "/path/to/the/code.py" | web_git_url(STORY.fmf_id.url, STORY.fmf_id.ref) }}

        {{ "/tmt/base.py" | web_git_url("https://github.com/teemtee/tmt.git", "main") }}
        -> https://github.com/teemtee/tmt/tree/main/tmt/base.py

    """

    path = Path(path_str) if path_str else None
    return web_git_url(url, ref, path)


def _template_filter_shell_quote(  # type: ignore[reportUnusedFunction,unused-ignore]
    s: str,
) -> str:
    """
    Return a shell-escaped version of the string.

    .. code-block:: jinja

        # "foo bar" -> "'foo bar'"
        {{ "foo bar" | shell_quote }}
    """

    return shlex.quote(s)


TEMPLATE_FILTERS: dict[str, Callable[..., Any]] = {
    _name.replace('_template_filter_', ''): _obj
    for _name, _obj in locals().items()
    if callable(_obj) and _name.startswith('_template_filter_')
}


def _template_test_unit(value: Any) -> bool:  # type: ignore[reportUnusedFunction,unused-ignore]
    """
    Return true if the object is a unit.

    .. code-block:: jinja

        {% if value is unit %}
            Value is a Pint's ``Quantity`` instance.
        {% endif %}
    """

    from pint import Quantity

    return isinstance(value, Quantity)


TEMPLATE_TESTS: dict[str, Callable[..., Any]] = {
    _name.replace('_template_test_', ''): _obj
    for _name, _obj in locals().items()
    if callable(_obj) and _name.startswith('_template_test_')
}


def default_template_environment() -> jinja2.Environment:
    """
    Create a Jinja2 environment with default settings.

    Adds common filters, and enables block trimming and left strip.
    """

    # S701: `autoescape=False` is dangerous and can lead to XSS.
    # As there can be many different template file formats, used to render various formats,
    # we need to explicitly set autoescape=False, as default might change in the future.
    # Potential improvements are being tracked in /teemtee/tmt/issues/2873

    environment = jinja2.Environment()  # noqa: S701

    environment.filters.update(TEMPLATE_FILTERS)
    environment.tests.update(TEMPLATE_TESTS)

    environment.trim_blocks = True
    environment.lstrip_blocks = True

    return environment


def render_template(
    template: str,
    template_filepath: Optional[Path] = None,
    environment: Optional[jinja2.Environment] = None,
    **variables: Any,
) -> str:
    """
    Render a template.

    :param template: template to render.
    :param template_filepath: path to the template file, if any.
    :param environment: Jinja2 environment to use.
    :param variables: variables to pass to the template.
    """

    environment = environment or default_template_environment()

    def raise_error(message: str) -> None:
        """
        An in-template helper for raising exceptions
        """

        raise Exception(message)

    if 'raise_error' not in variables:
        variables['raise_error'] = raise_error

    try:
        return environment.from_string(template).render(**variables).strip()

    except jinja2.exceptions.TemplateSyntaxError as exc:
        if template_filepath:
            raise GeneralError(
                f"Could not parse template '{template_filepath}' at line {exc.lineno}."
            ) from exc
        raise GeneralError(f"Could not parse template at line {exc.lineno}.") from exc

    except jinja2.exceptions.TemplateError as exc:
        if template_filepath:
            raise GeneralError(f"Could not render template '{template_filepath}'.") from exc
        raise GeneralError("Could not render template.") from exc


def render_template_file(
    template_filepath: Path,
    environment: Optional[jinja2.Environment] = None,
    **variables: Any,
) -> str:
    """
    Render a template from a file.

    :param template_filepath: path to the template file.
    :param environment: Jinja2 environment to use.
    :param variables: variables to pass to the template.
    """

    try:
        return render_template(
            template_filepath.read_text(), template_filepath, environment, **variables
        )

    except FileNotFoundError as exc:
        raise GeneralError(f"Could not open template '{template_filepath}'.") from exc



from __future__ import annotations

import sys

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

__all__ = [
    "deprecated",
]



"""
Main top-level command implementations.

.. note::

    We are currently refactoring :py:mod:`tmt.cli` into smaller modules.
    As of now, this module contains vast majority of tmt commands, but
    those will be continuously extracted until only the top-level
    command remains.
"""

from typing import TYPE_CHECKING, Any, Optional

import click
import fmf
import fmf.utils
from click import echo

import tmt
import tmt.base
import tmt.cli
import tmt.config
import tmt.convert
import tmt.identifier
import tmt.log
import tmt.options
import tmt.plugins
import tmt.plugins.plan_shapers
import tmt.steps
import tmt.templates
import tmt.utils
import tmt.utils.jira
from tmt.cli import CliInvocation, Context, ContextObject, CustomGroup, pass_context
from tmt.options import Deprecated, create_options_decorator, option
from tmt.utils import Command, Path, effective_workdir_root
from tmt.utils.themes import style

if TYPE_CHECKING:
    import tmt.steps.discover
    import tmt.steps.execute
    import tmt.steps.finish
    import tmt.steps.prepare
    import tmt.steps.provision
    import tmt.steps.report


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Common Options
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
verbosity_options = create_options_decorator(tmt.options.VERBOSITY_OPTIONS)
dry_options = create_options_decorator(tmt.options.DRY_OPTIONS)
force_dry_options = create_options_decorator(tmt.options.FORCE_DRY_OPTIONS)
again_option = create_options_decorator(tmt.options.AGAIN_OPTION)
fix_options = create_options_decorator(tmt.options.FIX_OPTIONS)
feeling_safe_option = create_options_decorator(tmt.options.FEELING_SAFE_OPTION)
workdir_root_options = create_options_decorator(tmt.options.WORKDIR_ROOT_OPTIONS)
filtering_options = create_options_decorator(tmt.options.FILTERING_OPTIONS)
filtering_options_long = create_options_decorator(tmt.options.FILTERING_OPTIONS_LONG)
filter_option = create_options_decorator(tmt.options.FILTER_OPTION)
fmf_source_options = create_options_decorator(tmt.options.FMF_SOURCE_OPTIONS)
story_flags_filter_options = create_options_decorator(tmt.options.STORY_FLAGS_FILTER_OPTIONS)
remote_plan_options = create_options_decorator(tmt.options.REMOTE_PLAN_OPTIONS)
lint_options = create_options_decorator(tmt.options.LINT_OPTIONS)
environment_options = create_options_decorator(tmt.options.ENVIRONMENT_OPTIONS)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Main
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@click.group(invoke_without_command=True, cls=CustomGroup)
@pass_context
@option(
    '-r',
    '--root',
    metavar='PATH',
    show_default=True,
    default='.',
    help="Path to the metadata tree root, '.' used by default.",
)
@option(
    '-c',
    '--context',
    metavar='DATA',
    multiple=True,
    help="""
         Set the fmf context. Use KEY=VAL or KEY=VAL1,VAL2... format to define individual
         dimensions or the @FILE notation to load data from provided yaml file. Can be specified
         multiple times.
         """,
)
@verbosity_options
@feeling_safe_option
@option(
    '--show-time',
    is_flag=True,
    help='If set, logging messages on the terminal would contain timestamps.',
)
@option(
    '--version',
    is_flag=True,
    help='Show tmt version and commit hash.',
)
@option(
    '--no-color',
    is_flag=True,
    default=False,
    help='Forces tmt to not use any colors in the output or logging.',
)
@option(
    '--force-color',
    is_flag=True,
    default=False,
    help='Forces tmt to use colors in the output and logging.',
)
@option(
    '--pre-check',
    is_flag=True,
    default=False,
    hidden=True,
    help='Run pre-checks on the git root. (Used by pre-commit wrapper).',
)
def main(
    click_contex: Context,
    root: str,
    context: list[str],
    no_color: bool,
    force_color: bool,
    show_time: bool,
    pre_check: bool,
    **kwargs: Any,
) -> None:
    """
    Test Management Tool
    """

    # Let Click know about the output width - this affects mostly --help output.
    click_contex.max_content_width = tmt.utils.OUTPUT_WIDTH

    # Show current tmt version and exit
    if kwargs.get('version'):
        print(f"tmt version: {tmt.__version__}")
        raise SystemExit(0)

    apply_colors_output, apply_colors_logging = tmt.log.decide_colorization(no_color, force_color)

    logger = tmt.log.Logger.create(
        apply_colors_output=apply_colors_output,
        apply_colors_logging=apply_colors_logging,
        **kwargs,
    )
    logger.add_console_handler(show_timestamps=show_time)

    # Propagate color setting to Click as well.
    click_contex.color = apply_colors_output

    # ignore[reportConstantRedefinition]: it looks like a constant, but
    # this redefinition is expected and on purpose. `EXCEPTION_LOGGER`
    # starts with the bootstrap logger, but now we constructed a better
    # one, and we need to switch.
    #
    # ignore[unused-ignore]: mypy does not report this issue, and the
    # ignore fools mypy into reporting the waiver as unused.
    tmt.cli.EXCEPTION_LOGGER = logger  # type: ignore[reportConstantRedefinition,unused-ignore]

    # Save click context and fmf context for future use
    tmt.utils.Common.store_cli_invocation(click_contex)

    # Run pre-checks
    if pre_check:
        git_command = tmt.utils.Command('git', 'rev-parse', '--show-toplevel')
        git_root = git_command.run(cwd=None, logger=logger).stdout
        if not git_root:
            raise tmt.utils.RunError("git rev-parse did not produce a path", git_command, 0)
        git_root = git_root.strip()
        git_command = tmt.utils.Command(
            'git', 'ls-files', '--error-unmatch', f"{git_root}/{root}/.fmf/version"
        )
        git_command.run(cwd=None, logger=logger)

    # Initialize metadata tree (from given path or current directory)
    tree = tmt.Tree(logger=logger, path=Path(root))

    # TODO: context object details need checks
    click_contex.obj = ContextObject(
        cli_context=click_contex,
        logger=logger,
        common=tmt.utils.Common(logger=logger),
        fmf_context=tmt.utils.FmfContext.from_spec('cli', context, logger),
        steps=set(),
        tree=tree,
    )

    # Show overview of available tests, plans and stories
    if click_contex.invoked_subcommand is None:
        tmt.Test.overview(tree)
        tmt.Plan.overview(tree)
        tmt.Story.overview(tree)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Run
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@main.group(chain=True, invoke_without_command=True, cls=CustomGroup)
@pass_context
@option(
    '-i',
    '--id',
    'id_',
    help='Run id (name or directory path).',
    metavar="ID",
)
@option(
    '-l',
    '--last',
    help='Execute the last run once again.',
    is_flag=True,
)
@option(
    '-r',
    '--rm',
    '--remove',
    'remove',
    is_flag=True,
    help='Remove the workdir when test run is finished.',
)
@option(
    '-k',
    '--keep',
    is_flag=True,
    help="""
         Keep all files in the run workdir after testing is done (skip pruning during the finish
         step).
         """,
)
@option(
    '--scratch',
    is_flag=True,
    help='Remove the run workdir before executing to start from scratch.',
)
@option(
    '--follow',
    is_flag=True,
    help='Output the logfile as it grows.',
)
@option(
    '-a',
    '--all',
    help='Run all steps, customize some.',
    is_flag=True,
)
@option(
    '-u',
    '--until',
    choices=tmt.steps.STEPS,
    help='Enable given step and all preceding steps.',
)
@option(
    '-s',
    '--since',
    choices=tmt.steps.STEPS,
    help='Enable given step and all following steps.',
)
@option(
    '-A',
    '--after',
    choices=tmt.steps.STEPS,
    help='Enable all steps after the given one.',
)
@option(
    '-B',
    '--before',
    choices=tmt.steps.STEPS,
    help='Enable all steps before the given one.',
)
@option(
    '-S',
    '--skip',
    choices=tmt.steps.STEPS,
    help='Skip given step(s) during test run execution.',
    multiple=True,
)
@option(
    '--on-plan-error',
    choices=['quit', 'continue'],
    default='quit',
    help="""
         What to do when plan fails to finish. Quit by default, or continue with the next plan.
         """,
)
@environment_options
@workdir_root_options
@verbosity_options
@force_dry_options
@again_option
def run(context: Context, id_: Optional[str], workdir_root: Optional[Path], **kwargs: Any) -> None:
    """
    Run test steps.
    """

    # Initialize
    logger = context.obj.logger.descend(logger_name='run', extra_shift=0)
    logger.apply_verbosity_options(**kwargs)

    run = tmt.Run(
        id_=Path(id_) if id_ is not None else None,
        tree=context.obj.tree,
        cli_invocation=CliInvocation.from_context(context),
        workdir_root=effective_workdir_root(workdir_root),
        logger=logger,
    )
    context.obj.run = run


for plugin_class in tmt.plugins.plan_shapers._PLAN_SHAPER_PLUGIN_REGISTRY.iter_plugins():
    run = create_options_decorator(plugin_class.run_options())(run)


# Steps options
run.add_command(tmt.steps.discover.DiscoverPlugin.command())
run.add_command(tmt.steps.provision.ProvisionPlugin.command())
run.add_command(tmt.steps.prepare.PreparePlugin.command())
run.add_command(tmt.steps.execute.ExecutePlugin.command())
run.add_command(tmt.steps.report.ReportPlugin.command())
run.add_command(tmt.steps.finish.FinishPlugin.command())
run.add_command(tmt.steps.Login.command())
run.add_command(tmt.steps.Reboot.command())


@run.command(name='plans')
@pass_context
@option(
    '-n',
    '--name',
    'names',
    metavar='[REGEXP|.]',
    multiple=True,
    help="Regular expression to match plan name or '.' for current directory.",
)
@filter_option
@option(
    '-c',
    '--condition',
    'conditions',
    metavar="EXPR",
    multiple=True,
    help="Use arbitrary Python expression for filtering.",
)
@option(
    '--link',
    'links',
    metavar="RELATION:TARGET",
    multiple=True,
    help="""
         Filter by linked objects (regular expressions are supported for both relation and target).
         """,
)
@option(
    '--default',
    is_flag=True,
    help="Use default plans even if others are available.",
)
@verbosity_options
def run_plans(context: Context, **kwargs: Any) -> None:
    """
    Select plans which should be executed.

    Regular expression can be used to filter plans by name.
    Use '.' to select plans under the current working directory.
    """

    tmt.base.Plan.store_cli_invocation(context)


@run.command(name='tests')
@pass_context
@option(
    '-n',
    '--name',
    'names',
    metavar='[REGEXP|.]',
    multiple=True,
    help="Regular expression to match test name or '.' for current directory.",
)
@filter_option
@option(
    '-c',
    '--condition',
    'conditions',
    metavar="EXPR",
    multiple=True,
    help="Use arbitrary Python expression for filtering.",
)
@option(
    '--link',
    'links',
    metavar="RELATION:TARGET",
    multiple=True,
    help="""
         Filter by linked objects (regular expressions are supported for both relation and target).
         """,
)
@option(
    '--failed-only',
    is_flag=True,
    default=False,
    help="""
         Select only failed tests from a previous run.
         Used when rerunning existing runs and requires either --id or --last option.
         """,
)
@verbosity_options
def run_tests(context: Context, **kwargs: Any) -> None:
    """
    Select tests which should be executed.

    Regular expression can be used to filter tests by name.
    Use '.' to select tests under the current working directory.
    """

    tmt.base.Test.store_cli_invocation(context)


# TODO: commands is unknown, needs revisit
@run.result_callback()
@pass_context
def finito(
    click_context: Context,
    /,
    commands: Any,
    *args: Any,
    **kwargs: Any,
) -> None:
    """
    Run tests if run defined
    """

    if click_context.obj.run:
        click_context.obj.run.go()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Test
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@main.group(invoke_without_command=True, cls=CustomGroup)
@pass_context
@verbosity_options
def tests(context: Context, /, **kwargs: Any) -> None:
    """
    Manage tests (L1 metadata).

    Check available tests, inspect their metadata.
    Convert old metadata into the new fmf format.
    """

    context.obj.logger = context.obj.logger.clone().apply_verbosity_options(**kwargs)

    # Show overview of available tests
    if context.invoked_subcommand is None:
        tmt.Test.overview(context.obj.tree)


@tests.command(name='ls')
@pass_context
@filtering_options
@verbosity_options
def tests_ls(context: Context, /, **kwargs: Any) -> None:
    """
    List available tests.

    Regular expression can be used to filter tests by name.
    Use '.' to select tests under the current working directory.
    """

    tmt.Test.store_cli_invocation(context)
    for test in context.obj.tree.tests():
        test.ls()


@tests.command(name='show')
@pass_context
@filtering_options
@verbosity_options
def tests_show(context: Context, /, **kwargs: Any) -> None:
    """
    Show test details.

    Regular expression can be used to filter tests by name.
    Use '.' to select tests under the current working directory.
    """

    tmt.Test.store_cli_invocation(context)

    logger = context.obj.logger.clone().apply_verbosity_options(**kwargs)

    for test in context.obj.tree.tests(logger=logger):
        test.show()
        echo()


_script_templates = fmf.utils.listed(tmt.templates.MANAGER.templates['script'], join='or')

_metadata_templates = fmf.utils.listed(tmt.templates.MANAGER.templates['test'], join='or')


@tests.command(name='create')
@pass_context
@click.argument('names', nargs=-1, metavar='[NAME]...')
@option(
    '-t',
    '--template',
    metavar='TEMPLATE',
    help=f'Test metadata template ({_metadata_templates}).',
    prompt=f'Test template ({_metadata_templates})',
)
@option(
    '-s',
    '--script',
    metavar='TEMPLATE',
    help=f'Test script template ({_script_templates}).',
)
@option(
    '--link',
    metavar='[RELATION:]TARGET',
    multiple=True,
    help='Link created test to the relevant issues.',
)
@verbosity_options
@force_dry_options
def tests_create(
    context: Context,
    names: list[str],
    template: str,
    script: Optional[str],
    force: bool,
    **kwargs: Any,
) -> None:
    """
    Create a new test based on given template.

    Specify directory name or use '.' to create tests under the
    current working directory.
    """

    assert context.obj.tree.root is not None  # narrow type
    tmt.Test.store_cli_invocation(context)
    tmt.Test.create(
        names=names,
        template=template,
        path=context.obj.tree.root,
        script=script,
        force=force,
        logger=context.obj.logger,
    )


@tests.command(name='import')
@pass_context
@click.argument('paths', nargs=-1, metavar='[PATH]...')
@option(
    '--nitrate / --no-nitrate',
    default=True,
    show_default=True,
    is_flag=True,
    help='Import test metadata from Nitrate.',
)
@option(
    '--polarion / --no-polarion',
    default=False,
    show_default=True,
    is_flag=True,
    help='Import test metadata from Polarion.',
)
@option(
    '--purpose / --no-purpose',
    default=True,
    show_default=True,
    is_flag=True,
    help='Migrate description from PURPOSE file.',
)
@option(
    '--makefile / --no-makefile',
    default=True,
    show_default=True,
    is_flag=True,
    help='Convert Beaker Makefile metadata.',
)
@option(
    '--restraint / --no-restraint',
    default=False,
    show_default=True,
    is_flag=True,
    help='Convert restraint metadata file.',
)
@option(
    '--general / --no-general',
    default=True,
    is_flag=True,
    help="""
         Detect components from linked nitrate general plans (overrides Makefile/restraint
         component).
         """,
)
@option(
    '--polarion-case-id',
    multiple=True,
    help="""
         Polarion Test case ID(s) to import data from. Can be provided multiple times. Can provide
         also test case name like: TEST-123:test_name
         """,
)
@option(
    '--link-polarion / --no-link-polarion',
    default=False,
    show_default=True,
    is_flag=True,
    help='Add Polarion link to fmf testcase metadata.',
)
@option(
    '--type',
    'types',
    metavar='TYPE',
    default=['multihost'],
    multiple=True,
    show_default=True,
    help="""
         Convert selected types from Makefile into tags. Use 'all' to convert all detected types.
         """,
)
@option(
    '--disabled',
    default=False,
    is_flag=True,
    help='Import disabled test cases from Nitrate as well.',
)
@option(
    '--manual',
    default=False,
    is_flag=True,
    help='Import manual test cases from Nitrate.',
)
@option(
    '--plan',
    metavar='PLAN',
    type=int,
    help='Identifier of test plan from which to import manual test cases.',
)
@option(
    '--case',
    metavar='CASE',
    type=int,
    help='Identifier of manual test case to be imported.',
)
@option(
    '--with-script',
    default=False,
    is_flag=True,
    help='Import manual cases with non-empty script field in Nitrate.',
)
@verbosity_options
@force_dry_options
def tests_import(
    context: Context,
    paths: list[str],
    makefile: bool,
    restraint: bool,
    general: bool,
    types: list[str],
    nitrate: bool,
    polarion: bool,
    polarion_case_id: list[str],
    link_polarion: bool,
    purpose: bool,
    disabled: bool,
    manual: bool,
    plan: int,
    case: int,
    with_script: bool,
    dry: bool,
    **kwargs: Any,
) -> None:
    """
    Import old test metadata into the new fmf format.

    Accepts one or more directories where old metadata are stored.
    By default all available sources and current directory are used.
    The following test metadata are converted for each source:

    \b
    makefile ..... summary, component, duration, require
    restraint .... name, description, entry_point, owner, max_time, repoRequires
    purpose ...... description
    nitrate ...... contact, component, tag, environment, relevancy, enabled
    polarion ..... summary, enabled, assignee, id, component, tag, description, link
    """

    tmt.Test.store_cli_invocation(context)

    if manual:
        if not (case or plan):
            raise tmt.utils.GeneralError(
                "Option --case or --plan is mandatory when using --manual."
            )
        tmt.convert.read_manual(plan, case, disabled, with_script, context.obj.logger)
        return

    if not paths:
        paths = ['.']
    for _path in paths:
        path = Path(_path)
        # Make sure we've got a real directory
        path = path.resolve()
        if not path.is_dir():
            raise tmt.utils.GeneralError(f"Path '{path}' is not a directory.")
        # Gather old metadata and store them as fmf
        common, individual = tmt.convert.read(
            path,
            makefile,
            restraint,
            nitrate,
            polarion,
            polarion_case_id,
            link_polarion,
            purpose,
            disabled,
            types,
            general,
            dry,
            context.obj.logger,
        )
        # Add path to common metadata if there are virtual test cases
        if individual:
            # TODO: fmf is not annotated yet, fmf.Tree.root is seen by pyright as possibly
            # `Unknown` type, therefore we need to help a bit.
            node = fmf.Tree(str(path))
            assert isinstance(node.root, str)  # narrow type
            root = Path(node.root)
            common['path'] = str(Path('/') / path.relative_to(root))
        # Store common metadata
        file_name = common.get('filename', 'main.fmf')
        common_path = path / file_name
        if not dry:
            tmt.convert.write(common_path, common)
        else:
            echo(style(f"Metadata would be stored into '{common_path}'.", fg='magenta'))
        # Store individual data (as virtual tests)
        for testcase in individual:
            if nitrate and testcase.get('extra-nitrate'):
                testcase_path = path / f'{testcase["extra-nitrate"]}.fmf'
            else:
                file_name = testcase.get('filename')
                if not file_name:
                    raise tmt.utils.ConvertError(
                        'Filename was not found, please set one with --polarion-case-id.'
                    )
                testcase_path = path / file_name
            if not dry:
                tmt.convert.write(testcase_path, testcase)
            else:
                echo(style(f"Metadata would be stored into '{testcase_path}'.", fg='magenta'))
        # Adjust runtest.sh content and permission if needed
        if not dry:
            tmt.convert.adjust_runtest(path / 'runtest.sh')


_test_export_formats = list(tmt.Test.get_export_plugin_registry().iter_plugin_ids())
_test_export_default = 'yaml'


@tests.command(name='export')
@pass_context
@filtering_options_long
@option(
    '-h',
    '--how',
    default=_test_export_default,
    show_default=True,
    help='Output format.',
    choices=_test_export_formats,
)
@option(
    '--format',
    default=_test_export_default,
    show_default=True,
    help='Output format.',
    deprecated=Deprecated('1.21', hint='use --how instead'),
    choices=_test_export_formats,
)
@option(
    '--nitrate',
    is_flag=True,
    help="Export test metadata to Nitrate.",
    deprecated=Deprecated('1.21', hint="use '--how nitrate' instead"),
)
@option(
    '--project-id',
    help='Use specific Polarion project ID.',
)
@option(
    '--link-polarion / --no-link-polarion',
    default=False,
    is_flag=True,
    help='Add Polarion link to fmf testcase metadata',
)
@option(
    '--bugzilla',
    is_flag=True,
    help="""
         Link Nitrate case to Bugzilla specified in the 'link' attribute with the relation
         'verifies'.
         """,
)
@option(
    '--ignore-git-validation',
    is_flag=True,
    help="""
         Ignore unpublished git changes and export to Nitrate. The case might not be able to be
         scheduled!
         """,
)
@option(
    '--append-summary / --no-append-summary',
    default=False,
    is_flag=True,
    help="""
         Include test summary in the Nitrate/Polarion test case summary as well. By default, only
         the repository name and test name are used.
         """,
)
@option(
    '--create',
    is_flag=True,
    help="Create test cases in nitrate if they don't exist.",
)
@option(
    '--general / --no-general',
    default=False,
    is_flag=True,
    help="""
         Link Nitrate case to component's General plan. Disabled by default. Note that this will
         unlink any previously connected general plans.
         """,
)
@option(
    '--link-runs / --no-link-runs',
    default=False,
    is_flag=True,
    help="""
         Link Nitrate case to all open runs of descendant plans of General plan. Disabled by
         default. Implies --general option.
         """,
)
@option(
    '--fmf-id',
    is_flag=True,
    help='Show fmf identifiers instead of test metadata.',
)
@option(
    '--duplicate / --no-duplicate',
    default=False,
    show_default=True,
    is_flag=True,
    help="""
         Allow or prevent creating duplicates in Nitrate by searching for existing test cases with
         the same fmf identifier.
         """,
)
@option(
    '-n',
    '--dry',
    is_flag=True,
    help="Run in dry mode. No changes, please.",
)
@option(
    '-d',
    '--debug',
    is_flag=True,
    help='Provide as much debugging details as possible.',
)
# TODO: move to `template` export plugin options
@option(
    '--template',
    metavar='PATH',
    help="Path to a template to use for rendering the export. Used with '--how=template' only.",
)
def tests_export(
    context: Context,
    format: str,
    how: str,
    nitrate: bool,
    bugzilla: bool,
    template: Optional[str],
    **kwargs: Any,
) -> None:
    """
    Export test data into the desired format.

    Regular expression can be used to filter tests by name.
    Use '.' to select tests under the current working directory.
    """

    tmt.Test.store_cli_invocation(context)

    if nitrate:
        context.obj.common.warn(
            "Option '--nitrate' is deprecated, please use '--how nitrate' instead."
        )
        how = 'nitrate'

    if format != _test_export_default:
        context.obj.common.warn("Option '--format' is deprecated, please use '--how' instead.")

        how = format

    # TODO: move this "requires bugzilla" flag to export plugin level.
    if bugzilla and how not in ('nitrate', 'polarion'):
        raise tmt.utils.GeneralError(
            "The --bugzilla option is supported only with --nitrate or --polarion for now."
        )

    if kwargs.get('fmf_id'):
        echo(
            tmt.base.FmfId.export_collection(
                collection=[test.fmf_id for test in context.obj.tree.tests()],
                format=how,
                template=Path(template) if template else None,
            )
        )

    else:
        echo(
            tmt.Test.export_collection(
                collection=context.obj.tree.tests(),
                format=how,
                template=Path(template) if template else None,
            )
        )


@tests.command(name="id")
@pass_context
@filtering_options
@verbosity_options
@force_dry_options
def tests_id(context: Context, /, **kwargs: Any) -> None:
    """
    Generate a unique id for each selected test.

    A new UUID is generated for each test matching the provided
    filter and the value is stored to disk. Existing identifiers
    are kept intact.
    """

    tmt.Test.store_cli_invocation(context)
    for test in context.obj.tree.tests():
        tmt.identifier.id_command(context, test.node, "test", dry=kwargs["dry"])


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Plan
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@main.group(invoke_without_command=True, cls=CustomGroup)
@pass_context
@verbosity_options
@remote_plan_options
def plans(context: Context, /, **kwargs: Any) -> None:
    """
    Manage test plans (L2 metadata).

    \b
    Search for available plans.
    Explore detailed test step configuration.
    """

    context.obj.logger = context.obj.logger.clone().apply_verbosity_options(**kwargs)

    # Show overview of available plans
    if context.invoked_subcommand is None:
        tmt.Plan.overview(context.obj.tree)


@plans.command(name='ls')
@pass_context
@filtering_options
@verbosity_options
@remote_plan_options
def plans_ls(context: Context, /, **kwargs: Any) -> None:
    """
    List available plans.

    Regular expression can be used to filter plans by name.
    Use '.' to select plans under the current working directory.
    """

    tmt.Plan.store_cli_invocation(context)
    for plan in context.obj.tree.plans():
        plan.ls()


@plans.command(name='show')
@pass_context
@filtering_options
@environment_options
@verbosity_options
@remote_plan_options
def plans_show(context: Context, /, **kwargs: Any) -> None:
    """
    Show plan details.

    Regular expression can be used to filter plans by name.
    Use '.' to select plans under the current working directory.
    """

    tmt.Plan.store_cli_invocation(context)

    logger = context.obj.logger.clone().apply_verbosity_options(**kwargs)

    for plan in context.obj.tree.plans(logger=logger):
        plan.show()
        echo()


_plan_templates = fmf.utils.listed(tmt.templates.MANAGER.templates['plan'], join='or')


@plans.command(name='create')
@pass_context
@click.argument('names', nargs=-1, metavar='[NAME]...')
@option(
    '-t',
    '--template',
    metavar='TEMPLATE',
    help=f'Plan template ({_plan_templates}).',
    prompt=f'Template ({_plan_templates})',
)
@option(
    '--discover',
    metavar='YAML',
    multiple=True,
    help='Discover phase content in yaml format.',
)
@option(
    '--provision',
    metavar='YAML',
    multiple=True,
    help='Provision phase content in yaml format.',
)
@option(
    '--prepare',
    metavar='YAML',
    multiple=True,
    help='Prepare phase content in yaml format.',
)
@option(
    '--execute',
    metavar='YAML',
    multiple=True,
    help='Execute phase content in yaml format.',
)
@option(
    '--report',
    metavar='YAML',
    multiple=True,
    help='Report phase content in yaml format.',
)
@option(
    '--finish',
    metavar='YAML',
    multiple=True,
    help='Finish phase content in yaml format.',
)
@option(
    '--link',
    metavar='[RELATION:]TARGET',
    multiple=True,
    help='Link created plan to the relevant issues.',
)
@verbosity_options
@force_dry_options
def plans_create(
    context: Context,
    names: list[str],
    template: str,
    force: bool,
    **kwargs: Any,
) -> None:
    """
    Create a new plan based on given template.
    """

    assert context.obj.tree.root is not None  # narrow type
    tmt.Plan.store_cli_invocation(context)
    tmt.Plan.create(
        names=names,
        template=template,
        path=context.obj.tree.root,
        force=force,
        logger=context.obj.logger,
    )


_plan_export_formats = list(tmt.Plan.get_export_plugin_registry().iter_plugin_ids())
_plan_export_default = 'yaml'


@plans.command(name='export')
@pass_context
@filtering_options_long
@option(
    '-h',
    '--how',
    default=_plan_export_default,
    show_default=True,
    help='Output format.',
    choices=_plan_export_formats,
)
@option(
    '--format',
    default=_plan_export_default,
    show_default=True,
    help='Output format.',
    deprecated=Deprecated('1.21', hint='use --how instead'),
    choices=_plan_export_formats,
)
@option(
    '-d',
    '--debug',
    is_flag=True,
    help='Provide as much debugging details as possible.',
)
# TODO: move to `template` export plugin options
@option(
    '--template',
    metavar='PATH',
    help="Path to a template to use for rendering the export. Used with '--how=template' only.",
)
@environment_options
def plans_export(
    context: Context,
    how: str,
    format: str,
    template: Optional[str],
    **kwargs: Any,
) -> None:
    """
    Export plans into desired format.

    Regular expression can be used to filter plans by name.
    Use '.' to select plans under the current working directory.
    """

    tmt.Plan.store_cli_invocation(context)

    if format != _test_export_default:
        context.obj.common.warn("Option '--format' is deprecated, please use '--how' instead.")

        how = format

    echo(
        tmt.Plan.export_collection(
            collection=context.obj.tree.plans(),
            format=how,
            template=Path(template) if template else None,
        )
    )


@plans.command(name="id")
@pass_context
@filtering_options
@verbosity_options
@force_dry_options
def plans_id(context: Context, /, **kwargs: Any) -> None:
    """
    Generate a unique id for each selected plan.

    A new UUID is generated for each plan matching the provided
    filter and the value is stored to disk. Existing identifiers
    are kept intact.
    """

    tmt.Plan.store_cli_invocation(context)
    for plan in context.obj.tree.plans():
        tmt.identifier.id_command(context, plan.node, "plan", dry=kwargs["dry"])


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Story
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@main.group(invoke_without_command=True, cls=CustomGroup)
@pass_context
@verbosity_options
def stories(context: Context, /, **kwargs: Any) -> None:
    """
    Manage user stories.

    \b
    Check available user stories.
    Explore coverage (test, implementation, documentation).
    """

    context.obj.logger = context.obj.logger.clone().apply_verbosity_options(**kwargs)

    # Show overview of available stories
    if context.invoked_subcommand is None:
        tmt.Story.overview(context.obj.tree)


@stories.command(name='ls')
@pass_context
@filtering_options_long
@story_flags_filter_options
@verbosity_options
def stories_ls(
    context: Context,
    /,
    implemented: bool,
    verified: bool,
    documented: bool,
    covered: bool,
    unimplemented: bool,
    unverified: bool,
    undocumented: bool,
    uncovered: bool,
    **kwargs: Any,
) -> None:
    """
    List available stories.

    Regular expression can be used to filter stories by name.
    Use '.' to select stories under the current working directory.
    """

    tmt.Story.store_cli_invocation(context)
    for story in context.obj.tree.stories():
        if story._match(
            implemented,
            verified,
            documented,
            covered,
            unimplemented,
            unverified,
            undocumented,
            uncovered,
        ):
            story.ls()


@stories.command(name='show')
@pass_context
@filtering_options_long
@story_flags_filter_options
@verbosity_options
def stories_show(
    context: Context,
    /,
    implemented: bool,
    verified: bool,
    documented: bool,
    covered: bool,
    unimplemented: bool,
    unverified: bool,
    undocumented: bool,
    uncovered: bool,
    **kwargs: Any,
) -> None:
    """
    Show story details.

    Regular expression can be used to filter stories by name.
    Use '.' to select stories under the current working directory.
    """

    tmt.Story.store_cli_invocation(context)

    logger = context.obj.logger.clone().apply_verbosity_options(**kwargs)

    for story in context.obj.tree.stories(logger=logger):
        if story._match(
            implemented,
            verified,
            documented,
            covered,
            unimplemented,
            unverified,
            undocumented,
            uncovered,
        ):
            story.show()
            echo()


_story_templates = fmf.utils.listed(tmt.templates.MANAGER.templates['story'], join='or')


@stories.command(name='create')
@pass_context
@click.argument('names', nargs=-1, metavar='[NAME]...')
@option(
    '-t',
    '--template',
    metavar='TEMPLATE',
    prompt=f'Template ({_story_templates})',
    help=f'Story template ({_story_templates}).',
)
@option(
    '--link',
    metavar='[RELATION:]TARGET',
    multiple=True,
    help='Link created story to the relevant issues.',
)
@verbosity_options
@force_dry_options
def stories_create(
    context: Context,
    names: list[str],
    template: str,
    force: bool,
    **kwargs: Any,
) -> None:
    """
    Create a new story based on given template.
    """

    assert context.obj.tree.root is not None  # narrow type
    tmt.Story.store_cli_invocation(context)
    tmt.Story.create(
        names=names,
        template=template,
        path=context.obj.tree.root,
        force=force,
        logger=context.obj.logger,
    )


@stories.command(name='coverage')
@option('--docs', is_flag=True, help='Show docs coverage.')
@option('--test', is_flag=True, help='Show test coverage.')
@option('--code', is_flag=True, help='Show code coverage.')
@pass_context
@filtering_options_long
@story_flags_filter_options
@verbosity_options
def stories_coverage(
    context: Context,
    /,
    code: bool,
    test: bool,
    docs: bool,
    implemented: bool,
    verified: bool,
    documented: bool,
    covered: bool,
    unimplemented: bool,
    unverified: bool,
    undocumented: bool,
    uncovered: bool,
    **kwargs: Any,
) -> None:
    """
    Show code, test and docs coverage for given stories.

    Regular expression can be used to filter stories by name.
    Use '.' to select stories under the current working directory.
    """

    tmt.Story.store_cli_invocation(context)

    def headfoot(text: str) -> None:
        """
        Format simple header/footer
        """

        echo(style(text.rjust(4) + ' ', fg='blue'), nl=False)

    header = False
    total = code_coverage = test_coverage = docs_coverage = 0
    if not any([code, test, docs]):
        code = test = docs = True
    for story in context.obj.tree.stories():
        # Check conditions
        if not story._match(
            implemented,
            verified,
            documented,
            covered,
            unimplemented,
            unverified,
            undocumented,
            uncovered,
        ):
            continue
        # Show header once
        if not header:
            if code:
                headfoot('code')
            if test:
                headfoot('test')
            if docs:
                headfoot('docs')
            headfoot('story')
            echo()
            header = True
        # Show individual stats
        status = story.coverage(code, test, docs)
        total += 1
        code_coverage += status[0]
        test_coverage += status[1]
        docs_coverage += status[2]
    # Summary
    if not total:
        return
    if code:
        headfoot(f'{round(100 * code_coverage / total)}%')
    if test:
        headfoot(f'{round(100 * test_coverage / total)}%')
    if docs:
        headfoot(f'{round(100 * docs_coverage / total)}%')
    headfoot(f"from {fmf.utils.listed(total, 'story')}")
    echo()


_story_export_formats = list(tmt.Story.get_export_plugin_registry().iter_plugin_ids())
_story_export_default = 'yaml'


@stories.command(name='export')
@pass_context
@filtering_options_long
@story_flags_filter_options
@option(
    '-h',
    '--how',
    default=_story_export_default,
    show_default=True,
    help='Output format.',
    choices=_story_export_formats,
)
@option(
    '--format',
    default=_story_export_default,
    show_default=True,
    help='Output format.',
    deprecated=Deprecated('1.21', hint='use --how instead'),
    choices=_story_export_formats,
)
@option(
    '-d',
    '--debug',
    is_flag=True,
    help='Provide as much debugging details as possible.',
)
# TODO: move to `template` export plugin options
@option(
    '--template',
    metavar='PATH',
    help="""
        Path to a template to use for rendering the export.
        Used with '--how=rst|template' only.
        """,
)
def stories_export(
    context: Context,
    how: str,
    format: str,
    implemented: bool,
    verified: bool,
    documented: bool,
    covered: bool,
    unimplemented: bool,
    unverified: bool,
    undocumented: bool,
    uncovered: bool,
    template: Optional[str],
    **kwargs: Any,
) -> None:
    """
    Export selected stories into desired format.

    Regular expression can be used to filter stories by name.
    Use '.' to select stories under the current working directory.
    """

    tmt.Story.store_cli_invocation(context)

    if format != _test_export_default:
        context.obj.common.warn("Option '--format' is deprecated, please use '--how' instead.")

        how = format

    stories = [
        story
        for story in context.obj.tree.stories(whole=True)
        if story._match(
            implemented,
            verified,
            documented,
            covered,
            unimplemented,
            unverified,
            undocumented,
            uncovered,
        )
    ]

    echo(
        tmt.Story.export_collection(
            collection=stories, format=how, template=Path(template) if template else None
        )
    )


@stories.command(name="id")
@pass_context
@filtering_options_long
@story_flags_filter_options
@verbosity_options
@force_dry_options
def stories_id(
    context: Context,
    /,
    implemented: bool,
    verified: bool,
    documented: bool,
    covered: bool,
    unimplemented: bool,
    unverified: bool,
    undocumented: bool,
    uncovered: bool,
    **kwargs: Any,
) -> None:
    """
    Generate a unique id for each selected story.

    A new UUID is generated for each story matching the provided
    filter and the value is stored to disk. Existing identifiers
    are kept intact.
    """

    tmt.Story.store_cli_invocation(context)
    for story in context.obj.tree.stories():
        if story._match(
            implemented,
            verified,
            documented,
            covered,
            unimplemented,
            unverified,
            undocumented,
            uncovered,
        ):
            tmt.identifier.id_command(context, story.node, "story", dry=kwargs["dry"])


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Clean
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Supported clean resources
CLEAN_RESOURCES: list[str] = ["guests", "runs", "images"]


@main.group(chain=True, invoke_without_command=True, cls=CustomGroup)
@pass_context
@option(
    '-l',
    '--last',
    is_flag=True,
    help='Clean resources related to the last run.',
)
@option(
    '-i',
    '--id',
    'id_',
    metavar="ID",
    multiple=True,
    help='Identifier (name or directory path) of the run to be cleaned.',
)
@option(
    '-k',
    '--keep',
    type=int,
    default=None,
    help='The number of latest workdirs to keep, clean the rest.',
)
@option(
    '-s',
    '--skip',
    choices=CLEAN_RESOURCES,
    help='The resources which should be kept on the disk.',
    multiple=True,
)
@workdir_root_options
@verbosity_options
@dry_options
def clean(
    context: Context,
    last: bool,
    id_: tuple[str, ...],
    keep: Optional[int],
    skip: list[str],
    workdir_root: Optional[Path],
    **kwargs: Any,
) -> None:
    """
    Clean workdirs, guests or images.

    Without any command, clean everything, stop the guests, remove
    all runs and then remove all images. Search for runs in
    /var/tmp/tmt, if runs are stored elsewhere, the path to them can
    be set using a subcommand (either runs or guests subcommand).

    The subcommands can be chained, the order of cleaning is always
    the same, irrespective of the order on the command line. First, all
    the guests are cleaned, followed by runs and images.
    """

    if last and id_:
        raise tmt.utils.GeneralError("Options --last and --id cannot be used together.")

    context.obj.clean_logger = context.obj.logger.descend(
        logger_name='clean', extra_shift=0
    ).apply_verbosity_options(**kwargs)

    echo(style('clean', fg='red'))
    clean_obj = tmt.Clean(
        logger=context.obj.clean_logger,
        parent=context.obj.common,
        cli_invocation=CliInvocation.from_context(context),
    )
    context.obj.clean = clean_obj
    exit_code = 0
    if context.invoked_subcommand is None:
        assert context.obj.clean_logger is not None  # narrow type
        workdir_root = effective_workdir_root(workdir_root)
        if not workdir_root.exists():
            raise tmt.utils.GeneralError(
                f"Path '{workdir_root}' does not exist, skipping guest, run and image cleanup."
            )
        # Create another level to the hierarchy so that logging indent is
        # consistent between the command and subcommands
        clean_obj = tmt.Clean(
            logger=context.obj.clean_logger.descend(
                logger_name='clean', extra_shift=0
            ).apply_verbosity_options(**kwargs),
            parent=clean_obj,
            cli_invocation=CliInvocation.from_context(context),
            workdir_root=workdir_root,
        )
        if 'guests' not in skip and not clean_obj.guests(id_, keep):
            exit_code = 1
        if 'runs' not in skip and not clean_obj.runs(id_, keep):
            exit_code = 1
        if 'images' not in skip and not clean_obj.images():
            exit_code = 1
        raise SystemExit(exit_code)


@clean.result_callback()
@pass_context
def perform_clean(
    click_context: Context,
    /,
    commands: Any,
    *args: Any,
    **kwargs: Any,
) -> None:
    """
    Perform clean actions in the correct order.

    We need to ensure that guests are always cleaned before the run workdirs
    even if the user specified them in reverse order.
    """

    clean_order = ("guests", "runs", "images")
    exit_code = 0
    for phase in clean_order:
        for partial in click_context.obj.clean_partials[phase]:
            if not partial():
                exit_code = 1
    raise SystemExit(exit_code)


@clean.command(name='runs')
@pass_context
@workdir_root_options
@option(
    '-l',
    '--last',
    is_flag=True,
    help='Clean the workdir of the last run.',
)
@option(
    '-i',
    '--id',
    'id_',
    metavar="ID",
    multiple=True,
    help='Run id(name or directory path) to clean workdir of. Can be specified multiple times.',
)
@option(
    '-k',
    '--keep',
    type=int,
    default=None,
    help='The number of latest workdirs to keep, clean the rest.',
)
@verbosity_options
@dry_options
def clean_runs(
    context: Context,
    workdir_root: Optional[Path],
    last: bool,
    id_: tuple[str, ...],
    keep: Optional[int],
    **kwargs: Any,
) -> None:
    """
    Clean workdirs of past runs.

    Remove all runs in '/var/tmp/tmt' by default.
    """

    defined = [last is True, bool(id_), keep is not None]
    if defined.count(True) > 1:
        raise tmt.utils.GeneralError("Options --last, --id and --keep cannot be used together.")
    if keep is not None and keep < 0:
        raise tmt.utils.GeneralError("--keep must not be a negative number.")
    if workdir_root and not workdir_root.exists():
        raise tmt.utils.GeneralError(f"Path '{workdir_root}' doesn't exist.")

    assert context.obj.clean_logger is not None  # narrow type

    clean_obj = tmt.Clean(
        logger=context.obj.clean_logger.descend(
            logger_name='clean-runs', extra_shift=0
        ).apply_verbosity_options(**kwargs),
        parent=context.obj.clean,
        cli_invocation=CliInvocation.from_context(context),
        workdir_root=effective_workdir_root(workdir_root),
    )
    context.obj.clean_partials["runs"].append(
        lambda: clean_obj.runs(
            (context.parent and context.parent.params.get('id_', [])) or id_,
            (context.parent and context.parent.params.get('keep', [])) or keep,
        )
    )


@clean.command(name='guests')
@pass_context
@workdir_root_options
@option(
    '-l',
    '--last',
    is_flag=True,
    help='Stop the guest of the last run.',
)
@option(
    '-i',
    '--id',
    'id_',
    metavar="ID",
    multiple=True,
    help='Run id(name or directory path) to stop the guest of. Can be specified multiple times.',
)
@option(
    '-k',
    '--keep',
    type=int,
    default=None,
    help='The number of latest guests to keep, clean the rest.',
)
@option(
    '-h',
    '--how',
    metavar='METHOD',
    help='Stop guests of the specified provision method.',
)
@verbosity_options
@dry_options
def clean_guests(
    context: Context,
    workdir_root: Optional[Path],
    last: bool,
    id_: tuple[str, ...],
    keep: Optional[int],
    **kwargs: Any,
) -> None:
    """
    Stop running guests of runs.

    Stop guests of all runs in '/var/tmp/tmt' by default.
    """

    if last and bool(id_):
        raise tmt.utils.GeneralError("Options --last and --id cannot be used together.")
    if workdir_root and not workdir_root.exists():
        raise tmt.utils.GeneralError(f"Path '{workdir_root}' doesn't exist.")

    assert context.obj.clean_logger is not None  # narrow type
    clean_obj = tmt.Clean(
        logger=context.obj.clean_logger.descend(
            logger_name='clean-guests', extra_shift=0
        ).apply_verbosity_options(**kwargs),
        parent=context.obj.clean,
        cli_invocation=CliInvocation.from_context(context),
        workdir_root=effective_workdir_root(workdir_root),
    )
    context.obj.clean_partials["guests"].append(
        lambda: clean_obj.guests(
            (context.parent and context.parent.params.get('id_', [])) or id_,
            (context.parent and context.parent.params.get('keep', [])) or keep,
        )
    )


@clean.command(name='images')
@pass_context
@workdir_root_options
@verbosity_options
@dry_options
def clean_images(context: Context, /, workdir_root: Optional[Path], **kwargs: Any) -> None:
    """
    Remove images of supported provision methods.

    Currently supported methods are:
     - testcloud
    """

    # FIXME: If there are more provision methods supporting this,
    #        we should add options to specify which provision should be
    #        cleaned, similarly to guests.
    assert context.obj.clean_logger is not None  # narrow type

    if workdir_root and not workdir_root.exists():
        raise tmt.utils.GeneralError(f"Path '{workdir_root}' doesn't exist.")

    clean_obj = tmt.Clean(
        logger=context.obj.clean_logger.descend(
            logger_name='clean-images', extra_shift=0
        ).apply_verbosity_options(**kwargs),
        parent=context.obj.clean,
        cli_invocation=CliInvocation.from_context(context),
        workdir_root=effective_workdir_root(workdir_root),
    )
    context.obj.clean_partials["images"].append(clean_obj.images)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Setup
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@main.group(cls=CustomGroup)
def setup(**kwargs: Any) -> None:
    """
    Setup the environment for working with tmt
    """


@setup.group(cls=CustomGroup)
def completion(**kwargs: Any) -> None:
    """
    Setup shell completions.

    By default, these commands only write a shell script to the output
    which can then be sourced from the shell's configuration file. Use
    the '--install' option to store and enable the configuration
    permanently.
    """


COMPLETE_VARIABLE = '_TMT_COMPLETE'
COMPLETE_SCRIPT = 'tmt-complete'


def setup_completion(shell: str, install: bool, context: Context, logger: tmt.log.Logger) -> None:
    """
    Setup completion based on the shell
    """

    config = tmt.config.Config(logger)
    # Fish gets installed into its special location where it is automatically
    # loaded.
    if shell == 'fish':
        script = Path('~/.config/fish/completions/tmt.fish').expanduser()
    # Bash and zsh get installed to tmt's config directory.
    else:
        script = Path(config.path) / f'{COMPLETE_SCRIPT}.{shell}'

    env_var = {COMPLETE_VARIABLE: f'{shell}_source'}

    logger = context.obj.logger

    completions = (
        Command('tmt')
        .run(env=tmt.utils.Environment.from_dict(env_var), cwd=None, logger=context.obj.logger)
        .stdout
    )
    if not completions:
        logger.warning("Unable to generate shell completion")
        return

    if install:
        Path(script).write_text(completions)
        # If requested, modify .bashrc or .zshrc
        if shell != 'fish':
            shell_config = Path(f'~/.{shell}rc').expanduser()
            shell_config.append_text('\n# Generated by tmt\n')
            shell_config.append_text(f'source {script}')

    else:
        logger.info(completions)


@completion.command(name='bash')
@pass_context
@option(
    '--install',
    '-i',
    'install',
    is_flag=True,
    help="""
         Persistently store the script to tmt's configuration directory and set it up by modifying
         '~/.bashrc'.
         """,
)
def completion_bash(context: Context, install: bool, **kwargs: Any) -> None:
    """
    Setup shell completions for bash
    """

    setup_completion('bash', install, context, context.obj.logger)


@completion.command(name='zsh')
@pass_context
@option(
    '--install',
    '-i',
    'install',
    is_flag=True,
    help="""
         Persistently store the script to tmt's configuration directory and set it up by modifying
         '~/.zshrc'.
         """,
)
def completion_zsh(context: Context, install: bool, **kwargs: Any) -> None:
    """
    Setup shell completions for zsh
    """

    setup_completion('zsh', install, context, context.obj.logger)


@completion.command(name='fish')
@pass_context
@option(
    '--install',
    '-i',
    'install',
    is_flag=True,
    help="Persistently store the script to '~/.config/fish/completions/tmt.fish'.",
)
def completion_fish(context: Context, install: bool, **kwargs: Any) -> None:
    """
    Setup shell completions for fish
    """

    setup_completion('fish', install, context, context.obj.logger)


@main.command(name='link')
@pass_context
@click.argument('names', nargs=-1, metavar='[TEST|PLAN|STORY]...')
@option(
    '--link',
    'links',
    metavar='[RELATION:]TARGET',
    multiple=True,
    help="""
        Issue to which tests, plans or stories should be linked.
        Can be provided multiple times.
        """,
)
@option(
    '--separate',
    is_flag=True,
    help="Create linking separately for multiple passed objects.",
)
def link(
    context: Context,
    names: list[str],
    links: list[str],
    separate: bool,
) -> None:
    """
    Create a link to tmt web service in an issue tracking software.

    Using the specified target, a link will be generated and added
    to an issue. Link is generated from names of tmt objects
    passed in arguments and configuration file.
    """

    tmt_objects = (
        context.obj.tree.tests(names=list(names))
        + context.obj.tree.plans(names=list(names))
        + context.obj.tree.stories(names=list(names))
    )

    if not tmt_objects:
        raise tmt.utils.GeneralError("No test, plan or story found for linking.")

    if not links:
        raise tmt.utils.GeneralError("Provide at least one link using the '--link' option.")

    for link in links:
        tmt.utils.jira.link(
            tmt_objects=tmt_objects,
            links=tmt.base.Links(data=link),
            separate=separate,
            logger=context.obj.logger,
        )



"""
Basic classes and code for tmt command line interface
"""

import collections
import enum
import functools
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, cast

import click
import fmf
import fmf.utils

import tmt
import tmt.base
import tmt.log
import tmt.plugins
import tmt.utils
import tmt.utils.rest
from tmt.container import container, simple_field

if TYPE_CHECKING:
    from tmt._compat.typing import Concatenate, ParamSpec

    P = ParamSpec('P')
    R = TypeVar('R')


#: A logger to use before the proper one can be established.
#:
#: .. warning::
#:
#:    This logger should be used with utmost care for logging while tmt
#:    is still starting. Once properly configured logger is spawned,
#:    honoring relevant options, this logger should not be used anymore.
_BOOTSTRAP_LOGGER = tmt.log.Logger.get_bootstrap_logger()

#: A logger to use for exception logging.
#:
#: .. warning::
#:
#:    This logger should be used with utmost care for logging exceptions
#:    only, no other traffic should be allowed. On top of that, the
#:    exception logging is handled by a dedicated function,
#:    :py:func:`tmt.utils.show_exception` - if you find yourself in need
#:    of logging an exception somewhere in the code, and you think about
#:    using this logger or calling ``show_exception()`` explicitly,
#:    it is highly likely you are not on the right track.
EXCEPTION_LOGGER: tmt.log.Logger = _BOOTSTRAP_LOGGER


# Explore available plugins (need to detect all supported methods first)
tmt.plugins.explore(_BOOTSTRAP_LOGGER)


class TmtExitCode(enum.IntEnum):
    # Quoting the specification:

    #: At least one test passed, there was no fail, warn or error.
    SUCCESS = 0

    #: There was a fail or warn identified, but no error.
    FAIL = 1

    #: Errors occurred during test execution.
    ERROR = 2

    #: No test results found.
    NO_RESULTS_FOUND = 3

    #: Tests were executed, and all reported the ``skip`` result.
    ALL_TESTS_SKIPPED = 4


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Click Context Object Container
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@container
class ContextObject:
    """
    Click Context Object container.

    In Click terms, this is "an arbitrary object of user data." In this container,
    tmt CLI code stores all structures relevant for the command execution. The
    container itself is then attached to :py:class:`click.Context` object Click
    manages across commands.
    """

    # "Parent" Click context
    cli_context: 'Context'

    logger: tmt.log.Logger
    common: tmt.utils.Common
    fmf_context: tmt.utils.FmfContext
    tree: tmt.Tree
    steps: set[str] = simple_field(default_factory=set[str])
    clean: Optional[tmt.Clean] = None
    clean_logger: Optional[tmt.log.Logger] = None
    clean_partials: collections.defaultdict[str, list[tmt.base.CleanCallback]] = simple_field(
        default_factory=lambda: cast(
            collections.defaultdict[str, list[tmt.base.CleanCallback]],
            collections.defaultdict(list),
        )
    )
    run: Optional[tmt.Run] = None


class Context(click.Context):
    """
    Custom :py:class:`click.Context`-like class for typing purposes.

    Objects of this class are never instantiated, it serves only as a type
    stub in commands below, to simplify handling and static analysis of
    ``context.obj``. There is no added functionality, the only change is
    a much narrower type of ``obj`` attribute.

    This class shall be used instead of the original :py:class:`click.Context`.
    Click is obviously not aware of our type annotations, and ``context``
    objects managed by Click would always be of type :py:class:`click.Context`,
    we would just convince mypy their ``obj`` attribute is no longer ``Any``.
    """

    # In contrast to the original Context, we *know* we do set obj to a valid
    # object, and every time we touch it, it should absolutely be not-None.
    obj: ContextObject

    max_content_width: Optional[int]


def pass_context(fn: 'Callable[Concatenate[Context, P], R]') -> 'Callable[P, R]':
    """
    Custom :py:func:`click.pass_context`-like decorator.

    Complementing the :py:class:`Context`, the goal of this decorator to
    announce the correct type of the ``context`` parameter. The original
    decorator annotates the parameter as ``click.Context``, but that is not
    what our command callables accept. So, on this boundary between tmt code
    and ``click`` API, we trick type checkers by isolating the necessary
    ``type: ignore[arg-type]``.
    """

    return click.pass_context(fn)  # type: ignore[arg-type]


@container
class CliInvocation:
    """
    A single CLI invocation of a tmt subcommand.

    Bundles together the Click context and options derived from it.
    A context alone might be good enough, but sometimes tmt needs to
    modify saved options. For custom command line options injected
    manually 'sources' is used to keep the parameter source.

    Serves as a clear boundary between invocations of classes
    representing various tmt subcommands and groups.
    """

    context: Optional[Context]
    options: dict[str, Any]

    @classmethod
    def from_context(cls, context: Context) -> 'CliInvocation':
        return CliInvocation(context=context, options=context.params)

    @classmethod
    def from_options(cls, options: dict[str, Any]) -> 'CliInvocation':
        """
        Inject custom options coming from the command line
        """

        invocation = CliInvocation(context=None, options=options)

        # ignore[reportGeneralTypeIssues]: pyright has troubles understanding it
        # *is* possible to assign to a cached property. Might an issue of our
        # simplified implementation.
        # ignore[unused-ignore]: silencing mypy's complaint about silencing
        # pyright's warning :)
        invocation.option_sources = dict.fromkeys(options, click.core.ParameterSource.COMMANDLINE)
        return invocation

    @functools.cached_property
    def option_sources(self) -> dict[str, click.core.ParameterSource]:
        if not self.context:
            return {}

        return self.context._parameter_source


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Custom Group
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class CustomGroup(click.Group):
    """
    Custom Click Group
    """

    # ignore[override]: expected, we want to use more specific `Context`
    # type than the one declared in superclass.
    def list_commands(self, context: Context) -> list[str]:  # type: ignore[override]
        """
        Prevent alphabetical sorting
        """

        return list(self.commands.keys())

    # ignore[override]: expected, we want to use more specific `Context`
    # type than the one declared in superclass.
    def get_command(  # type: ignore[override]
        self,
        context: Context,
        cmd_name: str,
    ) -> Optional[click.Command]:
        """
        Allow command shortening
        """

        # Backward-compatible 'test convert' (just temporary for now FIXME)
        cmd_name = cmd_name.replace('convert', 'import')
        # Support both story & stories
        cmd_name = cmd_name.replace('story', 'stories')
        found = click.Group.get_command(self, context, cmd_name)
        if found is not None:
            return found
        matches = [
            command for command in self.list_commands(context) if command.startswith(cmd_name)
        ]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, context, matches[0])
        context.fail(f"Did you mean {fmf.utils.listed(sorted(matches), join='or')}?")
        return None


class HelpFormatter(click.HelpFormatter):
    """
    Custom help formatter capable of rendering ReST syntax
    """

    # Override parent implementation
    def write_dl(
        self,
        rows: Sequence[tuple[str, str]],
        col_max: int = 30,
        col_spacing: int = 2,
    ) -> None:
        rows = [
            (option, tmt.utils.rest.render_rst(help, _BOOTSTRAP_LOGGER)) for option, help in rows
        ]

        super().write_dl(rows, col_max=col_max, col_spacing=col_spacing)


click.Context.formatter_class = HelpFormatter



from collections.abc import Mapping, Sequence
from typing import IO, Any, Optional, Union

import click.core
import click.testing


def reset_common() -> None:
    """
    Reset CLI invocation storage of classes derived from :py:class:`tmt.utils.Common`

    As CLI invocations are stored in class-level attributes, before each
    invocation of CLI in a test, we must reset these attributes to pretend the
    CLI is invoked for the very first time. Without this, after the very first
    invocation, subsequent CLI invocations would "inherit" options from the
    previous ones.

    A helper function to clear invocations of the "usual suspects". Classes that
    accept CLI options are reset.
    """

    from tmt.base import Core, Plan, Run, Story, Test, Tree
    from tmt.utils import Common, MultiInvokableCommon

    for klass in (Core, Run, Tree, Test, Plan, Story, Common, MultiInvokableCommon):
        klass.cli_invocation = None


class CliRunner(click.testing.CliRunner):
    def invoke(
        self,
        cli: click.core.BaseCommand,
        args: Optional[Union[str, Sequence[str]]] = None,
        input: Optional[Union[str, bytes, IO]] = None,
        env: Optional[Mapping[str, Optional[str]]] = None,
        catch_exceptions: bool = True,
        color: bool = False,
        **extra: Any,
    ) -> click.testing.Result:
        reset_common()

        from tmt.__main__ import import_cli_commands

        import_cli_commands()

        return super().invoke(
            cli,
            args=args,
            input=input,
            env=env,
            catch_exceptions=catch_exceptions,
            color=color,
            **extra,
        )



# mypy: disable-error-code="assignment"
from __future__ import annotations

import pydantic

if pydantic.__version__.startswith('1.'):
    from pydantic import (
        BaseModel,
        Extra,
        HttpUrl,
        ValidationError,
    )
else:
    from pydantic.v1 import (
        BaseModel,
        Extra,
        HttpUrl,
        ValidationError,
    )

__all__ = [
    "BaseModel",
    "Extra",
    "HttpUrl",
    "ValidationError",
]



from typing import Any, Optional

import tmt.log
from tmt.container import container, field
from tmt.steps.prepare.feature import PrepareFeatureData, ToggleableFeature, provides_feature
from tmt.steps.provision import Guest


@container
class EpelStepData(PrepareFeatureData):
    epel: Optional[str] = field(
        default=None,
        option='--epel',
        metavar='enabled|disabled',
        help='Whether EPEL repository should be installed & enabled or disabled.',
    )


@provides_feature('epel')
class Epel(ToggleableFeature):
    """
    Control Extra Packages for Enterprise Linux (EPEL) repository.

    `EPEL`__ is an initiative within the Fedora Project to provide high
    quality additional packages for CentOS Stream and Red Hat Enterprise
    Linux (RHEL).

    Enable or disable EPEL repository on the guest:

    .. code-block:: yaml

        prepare:
            how: feature
            epel: enabled

    .. code-block:: shell

        prepare --how feature --epel enabled

    __ https://docs.fedoraproject.org/en-US/epel/
    """

    _data_class = EpelStepData

    PLAYBOOKS = {'epel-enable.yaml', 'epel-disable.yaml'}

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def enable(cls, guest: Guest, logger: tmt.log.Logger) -> None:
        cls._run_playbook('enable', "epel-enable.yaml", guest, logger)

    @classmethod
    def disable(cls, guest: Guest, logger: tmt.log.Logger) -> None:
        cls._run_playbook('disable', "epel-disable.yaml", guest, logger)



import collections
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from tmt.base import Plan, Test
    from tmt.options import ClickOptionDecoratorType
    from tmt.steps.discover import TestOrigin

from tmt.plugins.plan_shapers import PlanShaper, provides_plan_shaper


@provides_plan_shaper('repeat')
class RepeatPlanShaper(PlanShaper):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def run_options(cls) -> list['ClickOptionDecoratorType']:
        from tmt.options import option

        return [
            option('--repeat', metavar='N', help='Repeat a plan N times.', type=int, default=-1)
        ]

    _inspected_plan_names: list[str] = []

    @classmethod
    def check(cls, plan: 'Plan', tests: list['TestOrigin']) -> bool:
        if not plan.my_run:
            return False

        repetitions = plan.my_run.opt('repeat')

        if repetitions <= 0:
            return False

        if plan.name in cls._inspected_plan_names:
            return False

        return True

    @classmethod
    def apply(cls, plan: 'Plan', tests: list['TestOrigin']) -> Iterator['Plan']:
        assert plan.my_run is not None

        if plan.name in cls._inspected_plan_names:
            return

        repetitions = plan.my_run.opt('repeat')

        plan.info(f'Repeating plan {repetitions} times.')

        batch: dict[str, list[Test]] = collections.defaultdict(list)

        for test_origin in tests:
            batch[test_origin.phase].append(test_origin.test)

        cls._inspected_plan_names.append(plan.name)

        for batch_id in range(1, repetitions + 1):
            derived_plan = plan.derive_plan(batch_id, batch)

            cls._inspected_plan_names.append(derived_plan.name)

            yield derived_plan



import datetime
import re
from re import Pattern
from typing import TYPE_CHECKING, Optional

import tmt.log
import tmt.steps.provision
import tmt.utils
import tmt.utils.themes
from tmt.checks import Check, CheckEvent, CheckPlugin, _RawCheck, provides_check
from tmt.container import container, field
from tmt.result import CheckResult, ResultOutcome
from tmt.steps.provision import GuestCapability
from tmt.utils import (
    Path,
    format_timestamp,
    render_command_report,
)

if TYPE_CHECKING:
    import tmt.base
    from tmt.steps.execute import TestInvocation
    from tmt.steps.provision import Guest

TEST_POST_DMESG_FILENAME = 'dmesg-{event}.txt'

DEFAULT_FAILURE_PATTERNS = [
    re.compile(pattern)
    for pattern in [
        r'Call Trace:',
        r'\ssegfault\s',
    ]
]


@container
class DmesgCheck(Check):
    failure_pattern: list[Pattern[str]] = field(
        default_factory=lambda: DEFAULT_FAILURE_PATTERNS[:],
        help="""
             List of regular expressions to look for in ``dmesg``
             output. If any of patterns is found, ``dmesg`` check will
             report ``fail`` result.
             """,
        metavar='PATTERN',
        normalize=tmt.utils.normalize_pattern_list,
        exporter=lambda patterns: [pattern.pattern for pattern in patterns],
        serialize=lambda patterns: [pattern.pattern for pattern in patterns],
        unserialize=lambda serialized: [re.compile(pattern) for pattern in serialized],
    )

    # TODO: fix `to_spec` of `Check` to support nested serializables
    def to_spec(self) -> _RawCheck:
        spec = super().to_spec()

        spec['failure-pattern'] = [  # type: ignore[reportGeneralTypeIssues,typeddict-unknown-key,unused-ignore]
            pattern.pattern for pattern in self.failure_pattern
        ]

        return spec

    def to_minimal_spec(self) -> _RawCheck:
        return self.to_spec()

    @classmethod
    def _fetch_dmesg(
        cls,
        guest: tmt.steps.provision.Guest,
        logger: tmt.log.Logger,
    ) -> tmt.utils.CommandOutput:
        def _test_output_logger(
            key: str,
            value: Optional[str] = None,
            color: tmt.utils.themes.Style = None,
            shift: int = 2,
            level: int = 3,
            topic: Optional[tmt.log.Topic] = None,
        ) -> None:
            logger.verbose(
                key=key, value=value, color=color, shift=shift, level=level, topic=topic
            )

        if guest.facts.has_capability(GuestCapability.SYSLOG_ACTION_READ_CLEAR):
            script = tmt.utils.ShellScript('dmesg -c')

        else:
            script = tmt.utils.ShellScript('dmesg')

        if not guest.facts.is_superuser:
            script = tmt.utils.ShellScript(f'sudo {script.to_shell_command()}')

        return guest.execute(script, log=_test_output_logger)

    def _save_dmesg(
        self, invocation: 'TestInvocation', event: CheckEvent, logger: tmt.log.Logger
    ) -> tuple[ResultOutcome, Path]:
        assert invocation.phase.step.workdir is not None  # narrow type

        timestamp = format_timestamp(datetime.datetime.now(datetime.timezone.utc))

        path = invocation.check_files_path / TEST_POST_DMESG_FILENAME.format(event=event.value)

        try:
            output = self._fetch_dmesg(invocation.guest, logger)

        except tmt.utils.RunError as exc:
            outcome = ResultOutcome.ERROR
            output = exc.output

        else:
            outcome = ResultOutcome.PASS

            if any(pattern.search(output.stdout or '') for pattern in self.failure_pattern):
                outcome = ResultOutcome.FAIL

        invocation.phase.write(
            path,
            '\n'.join(render_command_report(label=f'Acquired at {timestamp}', output=output)),
            mode='a',
        )

        return outcome, path.relative_to(invocation.phase.step.workdir)


@provides_check('dmesg')
class Dmesg(CheckPlugin[DmesgCheck]):
    #
    # This plugin docstring has been reviewed and updated to follow
    # our documentation best practices. When changing it, please make
    # sure new changes are following them as well.
    #
    # https://tmt.readthedocs.io/en/stable/contribute.html#docs
    #
    """
    Save the content of kernel ring buffer (aka "console") into a file.

    The check saves one file before the test, and then again
    when test finishes.

    .. code-block:: yaml

        check:
          - how: dmesg

    Check will identify patterns that signal kernel crashes and
    core dumps, and when detected, it will report as failed result.
    It is possible to define custom patterns:

    .. code-block:: yaml

        check:
          - how: dmesg
            failure-pattern:
              # These are default patterns
              - 'Call Trace:
              - '\\ssegfault\\s'

              # More patterns to look for
              - '\\[Firmware Bug\\]'

    .. versionadded:: 1.28

    .. versionchanged:: 1.33
       ``failure-pattern`` has been added.
    """

    _check_class = DmesgCheck

    @classmethod
    def essential_requires(
        cls,
        guest: 'Guest',
        test: 'tmt.base.Test',
        logger: tmt.log.Logger,
    ) -> list['tmt.base.DependencySimple']:
        if not guest.facts.has_capability(GuestCapability.SYSLOG_ACTION_READ_ALL):
            return []

        # Avoid circular imports
        import tmt.base

        return [tmt.base.DependencySimple('/usr/bin/dmesg')]

    @classmethod
    def before_test(
        cls,
        *,
        check: 'DmesgCheck',
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[CheckResult]:
        if not invocation.guest.facts.has_capability(GuestCapability.SYSLOG_ACTION_READ_ALL):
            return [CheckResult(name='dmesg', result=ResultOutcome.SKIP)]

        outcome, path = check._save_dmesg(invocation, CheckEvent.BEFORE_TEST, logger)

        return [CheckResult(name='dmesg', result=outcome, log=[path])]

    @classmethod
    def after_test(
        cls,
        *,
        check: 'DmesgCheck',
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[CheckResult]:
        if not invocation.guest.facts.has_capability(GuestCapability.SYSLOG_ACTION_READ_ALL):
            return [CheckResult(name='dmesg', result=ResultOutcome.SKIP)]

        if not invocation.is_guest_healthy:
            return [CheckResult(name='dmesg', result=ResultOutcome.SKIP)]

        outcome, path = check._save_dmesg(invocation, CheckEvent.AFTER_TEST, logger)

        return [CheckResult(name='dmesg', result=outcome, log=[path])]



import re
import sys

from tmt.__main__ import run_cli


def _run_precommit() -> None:  # pyright: ignore[reportUnusedFunction] (used by project.scripts)
    """
    A simple wrapper that re-orders cli arguments before :py:func:`run_cli`.

    This utility is needed in order to move arguments like ``--root`` before
    the ``lint`` keyword when run by ``pre-commit``.

    Only a limited number of arguments are reordered.
    """

    # Only re-ordering a few known/necessary cli arguments of `main` command
    # Could be improved if the click commands can be introspected like with
    # `attrs.fields`
    # https://github.com/pallets/click/issues/2709

    argv = sys.argv
    # Get the last argument that starts with `-`. Other inputs after that are
    # assumed to be filenames (which can be thousands of them)
    for last_arg in reversed(range(len(argv))):
        if argv[last_arg].startswith("-"):
            break
    else:
        # If there were no args passed, then just `run_cli`
        run_cli()
        return

    # At this point we need to check and re-order the arguments if needed
    for pattern, nargs in [
        (r"(--root$|-r)", 1),
        (r"(--root=.*)", 0),
        (r"(--verbose$|-v+)", 0),
        (r"(--debug$|-d+)", 0),
        (r"(--pre-check$)", 0),
    ]:
        if any(match := re.match(pattern, arg) for arg in argv[1 : last_arg + 1]):
            assert match
            # Actual argument matched
            actual_arg = match.group(1)
            index = argv.index(actual_arg)
            # Example of reorder args:
            # ["tmt", "tests", "lint", "--fix", "--root", "/some/path", "some_file"]
            # - argv[0]: name of the program is always first
            #   ["tmt"]
            # - argv[index:index+1+nargs]: args to move (pass to `tmt.cli._root.main`)
            #   ["--root", "/some/path"]
            # - argv[1:index]: all other keywords (`lint` keyword is in here)
            #   ["tests", "lint", "--fix"]
            # - argv[index+1+nargs:]: All remaining keywords
            #   ["some_file"]
            # After reorder:
            # ["tmt", "--root", "/some/path", "tests", "lint", "--fix", "some_file"]
            argv = (
                argv[0:1]
                + argv[index : index + 1 + nargs]
                + argv[1:index]
                + argv[index + 1 + nargs :]
            )
    # Finally move the reordered args to sys.argv and run the cli
    sys.argv = argv
    run_cli()


if __name__ == "__main__":
    _run_precommit()



import copy
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, cast

import click
import fmf

import tmt
import tmt.steps
from tmt.container import container
from tmt.options import option
from tmt.plugins import PluginRegistry
from tmt.result import PhaseResult
from tmt.steps import (
    Action,
    ActionTask,
    Method,
    PhaseQueue,
    PluginTask,
    PullTask,
    sync_with_guests,
)
from tmt.steps.provision import Guest

if TYPE_CHECKING:
    import tmt.cli


@container
class FinishStepData(tmt.steps.WhereableStepData, tmt.steps.StepData):
    pass


FinishStepDataT = TypeVar('FinishStepDataT', bound=FinishStepData)


class FinishPlugin(tmt.steps.Plugin[FinishStepDataT, list[PhaseResult]]):
    """
    Common parent of finish plugins
    """

    # ignore[assignment]: as a base class, FinishStepData is not included in
    # FinishStepDataT.
    _data_class = FinishStepData  # type: ignore[assignment]

    # Methods ("how: ..." implementations) registered for the same step.
    _supported_methods: PluginRegistry[Method] = PluginRegistry('step.finish')

    @classmethod
    def base_command(
        cls,
        usage: str,
        method_class: Optional[type[click.Command]] = None,
    ) -> click.Command:
        """
        Create base click command (common for all finish plugins)
        """

        # Prepare general usage message for the step
        if method_class:
            usage = Finish.usage(method_overview=usage)

        # Create the command
        @click.command(cls=method_class, help=usage)
        @click.pass_context
        @option('-h', '--how', metavar='METHOD', help='Use specified method for finishing tasks.')
        @tmt.steps.PHASE_OPTIONS
        def finish(context: 'tmt.cli.Context', **kwargs: Any) -> None:
            context.obj.steps.add('finish')
            Finish.store_cli_invocation(context)

        return finish

    def go(
        self,
        *,
        guest: 'Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        self.go_prolog(logger)

        return []


class Finish(tmt.steps.Step):
    """
    Perform the finishing tasks and clean up provisioned guests.

    Additional actions to be performed after the test execution has been
    completed. Counterpart of the ``prepare`` step useful for various
    cleanup actions. Also takes care of stopping and removing guests.

    Note that the ``finish`` step is also run when any of the previous
    steps failed (for example when the environment preparation was not
    successful) so that provisioned systems are not kept running.
    """

    _plugin_base_class = FinishPlugin

    @property
    def _preserved_workdir_members(self) -> set[str]:
        """
        A set of members of the step workdir that should not be removed.
        """

        return {*super()._preserved_workdir_members, 'results.yaml'}

    def wake(self) -> None:
        """
        Wake up the step (process workdir and command line)
        """

        super().wake()

        # Choose the right plugin and wake it up
        for data in self.data:
            # FIXME: cast() - see https://github.com/teemtee/tmt/issues/1599
            plugin = cast(FinishPlugin[FinishStepData], FinishPlugin.delegate(self, data=data))
            plugin.wake()
            # Add plugin only if there are data
            if not plugin.data.is_bare:
                self._phases.append(plugin)

        # Nothing more to do if already done and not asked to run again
        if self.status() == 'done' and not self.should_run_again:
            self.debug('Finish wake up complete (already done before).', level=2)
        # Save status and step data (now we know what to do)
        else:
            self.status('todo')
            self.save()

    def summary(self) -> None:
        """
        Give a concise summary
        """

        tasks = fmf.utils.listed(self.phases(), 'task')
        self.info('summary', f'{tasks} completed', 'green', shift=1)

    def go(self, force: bool = False) -> None:
        """
        Execute finishing tasks
        """

        super().go(force=force)

        # Nothing more to do if already done
        if self.status() == 'done':
            self.info('status', 'done', 'green', shift=1)
            self.summary()
            self.actions()
            return

        # Nothing to do if no guests were provisioned
        if not self.plan.provision.guests:
            self.warn("Nothing to finish, no guests provisioned.", shift=1)
            return

        if self.plan.provision.ready_guests:
            # Prepare guests
            guest_copies: list[Guest] = []

            for guest in self.plan.provision.ready_guests:
                # Create a guest copy and change its parent so that the
                # operations inside finish plugins on the guest use the
                # finish step config rather than provision step config.
                guest_copy = copy.copy(guest)
                guest_copy.inject_logger(
                    guest._logger.clone().apply_verbosity_options(**self._cli_options)
                )
                guest_copy.parent = self

                guest_copies.append(guest_copy)

            queue: PhaseQueue[FinishStepData, list[PhaseResult]] = PhaseQueue(
                'finish', self._logger.descend(logger_name=f'{self}.queue')
            )

            for phase in self.phases(classes=(Action, FinishPlugin)):
                if isinstance(phase, Action):
                    queue.enqueue_action(phase=phase)

                elif phase.enabled_by_when:
                    queue.enqueue_plugin(
                        phase=phase,  # type: ignore[arg-type]
                        guests=[guest for guest in guest_copies if phase.enabled_on_guest(guest)],
                    )

            failed_tasks: list[
                Union[ActionTask, PluginTask[FinishStepData, list[PhaseResult]]]
            ] = []
            results: list[PhaseResult] = []

            for outcome in queue.run():
                if not isinstance(outcome.phase, FinishPlugin):
                    continue

                if outcome.exc:
                    outcome.logger.fail(str(outcome.exc))

                    failed_tasks.append(outcome)
                    continue

                if outcome.result:
                    results += outcome.result

            self._save_results(results)

            if failed_tasks:
                raise tmt.utils.GeneralError(
                    'finish step failed',
                    causes=[outcome.exc for outcome in failed_tasks if outcome.exc is not None],
                )

            # To separate "finish" from "pull" queue visually
            self.info('')

            # Pull artifacts created in the plan data directory
            # if there was at least one plugin executed
            if self.phases() and guest_copies:
                sync_with_guests(
                    self,
                    'pull',
                    PullTask(
                        logger=self._logger, guests=guest_copies, source=self.plan.data_directory
                    ),
                    self._logger,
                )

                # To separate "finish" from "pull" queue visually
                self.info('')

            self.summary()

        # Stop and remove provisioned guests, even the partially provisioned ones.
        for guest in self.plan.provision.guests:
            guest.fetch_logs(logger=self._logger)
            guest.stop()
            guest.remove()

        # Prune all irrelevant files and dirs
        assert self.plan.my_run is not None
        if not (self.plan.my_run.opt('keep') or self.is_dry_run):
            self.plan.prune()

        # Update status and save
        self.status('done')
        self.save()
